import astropy.units as u
import numpy as np
from snewpy.models import presn
from snewpy.flavor_transformation import AdiabaticMSW, MassHierarchy
from snewpy.rate_calculator import RateCalculator
import pytest

pytestmark=pytest.mark.snowglobes

rc = RateCalculator()

distance = 200*u.pc
 #SNOwGLoBES detector for water Cerenkov
T = np.geomspace(-1*u.hour, -1*u.min,1000)
E = np.linspace(0,20,100)*u.MeV

@pytest.mark.parametrize('model_class',[presn.Odrzywolek_2010, presn.Kato_2017, presn.Patton_2017, presn.Yoshida_2016])
@pytest.mark.parametrize('transformation',[AdiabaticMSW(mh=mh) for mh in MassHierarchy])
@pytest.mark.parametrize('detector', ["wc100kt30prct"])
def test_presn_rate(model_class, transformation, detector):
    model = model_class(progenitor_mass=15*u.Msun)
    flux = model.get_flux(T, E, distance=distance, flavor_xform=transformation)
    rate = rc.run(flux, detector='scint20kt', detector_effects=False)['ibd']
    ibd_events = rate.integrate_or_sum('time').integrate_or_sum('energy').array.squeeze()
    assert 10<ibd_events<1000


# -*- coding: utf-8 -*-
"""Unit tests for the neutrino submodule.
"""
import unittest

from snewpy.neutrino import Flavor
from snewpy.flavor_transformation import NoTransformation
from snewpy.models.ccsn import Nakazato_2013, Tamborra_2014, OConnor_2013, OConnor_2015, \
                          Sukhbold_2015, Bollig_2016, Walk_2018, \
                          Walk_2019, Fornax_2019, Warren_2020, \
                          Kuroda_2020, Fornax_2021, Zha_2021, \
                          Fornax_2022, Mori_2023, Bugli_2021, Fischer_2020

from astropy import units as u

import numpy as np


class TestModels(unittest.TestCase):

    def test_param_combinations(self):
        models = [Nakazato_2013,
                  Tamborra_2014,
                  OConnor_2013,
                  OConnor_2015,
                  Sukhbold_2015,
                  Bollig_2016,
                  Walk_2018,
                  Walk_2019,
                  Fornax_2019,
                  Warren_2020,
                  Kuroda_2020,
                  Fornax_2021,
                  Zha_2021,
                  Fornax_2022,
                  Mori_2023,
                  ]

        for model in models:
            for pc in model.get_param_combinations():
                model(**pc)  # Initialize

    def test_Nakazato_2013(self):
        """
        Instantiate a set of 'Nakazato 2013' models
        """
        for z in [0.004, 0.02]:
            for trev in [100, 200, 300]:
                for mass in [13., 20., 50.]:
                    model = Nakazato_2013(progenitor_mass=mass * u.Msun, metallicity=z, eos='shen',
                                          revival_time=trev * u.ms)

                    self.assertEqual(model.metadata['EOS'], 'shen')
                    self.assertEqual(model.metadata['Progenitor mass'], mass*u.Msun)
                    self.assertEqual(model.metadata['Revival time'], trev*u.ms)
                    self.assertEqual(model.metadata['Metallicity'], z)

                    # Check that times are in proper units.
                    t = model.get_time()
                    self.assertTrue(t.unit, u.s)
                    self.assertEqual(model.time[0], -50*u.ms)

                    # Check that we can compute flux dictionaries.
                    f = model.get_initial_spectra(0*u.s, 10*u.MeV)
                    self.assertEqual(type(f), dict)
                    self.assertEqual(len(f), len(Flavor))
                    self.assertEqual(f[Flavor.NU_E].unit, 1/(u.erg * u.s))

    def test_Tamborra_2014(self):
        """
        Instantiate a set of 'Tamborra 2014' models
        """
        for mass in [11.2, 20., 27.]:
            for angles in [1,2,3]:
                if mass != 27. and angles > 1:
                    with self.assertRaises(ValueError):
                        model = Tamborra_2014(progenitor_mass=mass*u.Msun, eos='LS220', direction=angles)
                else:
                    model = Tamborra_2014(progenitor_mass=mass*u.Msun, eos='LS220',direction=angles)

                    self.assertEqual(model.metadata['EOS'], 'LS220')
                    self.assertEqual(model.metadata['Progenitor mass'], mass*u.Msun)

                    # Check that times are in proper units.
                    t = model.get_time()
                    self.assertTrue(t.unit, u.s)

                    # Check that we can compute flux dictionaries.
                    f = model.get_initial_spectra(0*u.s, 10*u.MeV)
                    self.assertEqual(type(f), dict)
                    self.assertEqual(len(f), len(Flavor))
                    self.assertEqual(f[Flavor.NU_E].unit, 1/(u.erg * u.s))

    def test_Bugli_2021(self):
        """
        Instantiate a set of 'Bugli 2021' models
        """
        for Bfield in ['hydro', 'L1', 'L2']:
            for direction in ['average','equator','north','south']:
                if Bfield == 'L2':
                    for grav in ['A','B']:
                        model = Bugli_2021(Bfield=Bfield, grav=grav, rotation=None, direction=direction)
                if Bfield == 'L1':
                    for rotation in [0,90]:
                        model = Bugli_2021(Bfield=Bfield, grav=None, rotation=rotation, direction=direction)
                if Bfield == 'hydro':
                    model = Bugli_2021(Bfield=Bfield, grav=None, rotation=None, direction=direction)

                self.assertEqual(model.metadata['EOS'], 'LS220')
                self.assertEqual(model.metadata['Progenitor mass'], 35*u.Msun)

                # Check that times are in proper units.
                t = model.get_time()
                self.assertTrue(t.unit, u.s)

                # Check that we can compute flux dictionaries.
                f = model.get_initial_spectra(0*u.s, 10*u.MeV)
                self.assertEqual(type(f), dict)
                self.assertEqual(len(f), len(Flavor))
                self.assertEqual(f[Flavor.NU_E].unit, 1/(u.erg * u.s))

    def test_OConnor_2013(self):
        """
        Instantiate a set of "O'Connor 2015" models
        """
        for mass in list(range(12, 34)) + list(range(35, 61, 5)) + [70, 80, 100, 120]:
            for eos in ['LS220', 'HShen']:
                model = OConnor_2013(progenitor_mass=mass*u.Msun, eos=eos)

                self.assertEqual(model.metadata['EOS'], eos)
                self.assertEqual(model.metadata['Progenitor mass'].value, mass)

                # Check that times are in proper units.
                t = model.get_time()
                self.assertTrue(t.unit, u.s)

                # Check that we can compute flux dictionaries.
                f = model.get_initial_spectra(0*u.s, 10*u.MeV)
                self.assertEqual(type(f), dict)
                self.assertEqual(len(f), len(Flavor))
                self.assertEqual(f[Flavor.NU_E].unit, 1/(u.erg * u.s))

    def test_OConnor_2015(self):
        """
        Instantiate a set of "O'Connor 2015" models
        """
        model = OConnor_2015(progenitor_mass=40*u.Msun, eos='LS220')

        self.assertEqual(model.metadata['EOS'], 'LS220')
        self.assertEqual(model.metadata['Progenitor mass'], 40*u.Msun)

        # Check that times are in proper units.
        t = model.get_time()
        self.assertTrue(t.unit, u.s)

        # Check that we can compute flux dictionaries.
        f = model.get_initial_spectra(0*u.s, 10*u.MeV)
        self.assertEqual(type(f), dict)
        self.assertEqual(len(f), len(Flavor))
        self.assertEqual(f[Flavor.NU_E].unit, 1/(u.erg * u.s))

    def test_Sukhbold_2015(self):
        """
        Instantiate a set of 'Sukhbold 2015' models
        """
        for mass in [9.6, 27.]:
            for eos in ['LS220', 'SFHo']:
                model = Sukhbold_2015(progenitor_mass=mass * u.Msun, eos=eos)

                self.assertEqual(model.metadata['EOS'], eos)
                self.assertEqual(model.metadata['Progenitor mass'].value, mass)

                # Check that times are in proper units.
                t = model.get_time()
                self.assertTrue(t.unit, u.s)

                # Check that we can compute flux dictionaries.
                f = model.get_initial_spectra(0*u.s, 10*u.MeV)
                self.assertEqual(type(f), dict)
                self.assertEqual(len(f), len(Flavor))
                self.assertEqual(f[Flavor.NU_E].unit, 1/(u.erg * u.s))

    def test_Bollig_2016(self):
        """
        Instantiate a set of 'Bollig 2016' models
        """
        for mass in [11.2, 27.]:
            model = Bollig_2016(progenitor_mass=mass*u.Msun, eos='LS220')

            self.assertEqual(model.metadata['EOS'], 'LS220')
            self.assertEqual(model.metadata['Progenitor mass'], mass*u.Msun)

            # Check that times are in proper units.
            t = model.get_time()
            self.assertTrue(t.unit, u.s)

            # Check that we can compute flux dictionaries.
            f = model.get_initial_spectra(0*u.s, 10*u.MeV)
            self.assertEqual(type(f), dict)
            self.assertEqual(len(f), len(Flavor))
            self.assertEqual(f[Flavor.NU_E].unit, 1/(u.erg * u.s))

    def test_Walk_2018(self):
        """
        Instantiate a set of 'Walk 2018' models
        """
        mass = 15.
        for angles in [1,2,3]:
            for rot in ['fast','slow','non']:
                model = Walk_2018(progenitor_mass=mass*u.Msun, direction=angles, rotation=rot)

                self.assertEqual(model.metadata['EOS'], 'LS220')
                self.assertEqual(model.metadata['Progenitor mass'], mass*u.Msun)

                # Check that times are in proper units.
                t = model.get_time()
                self.assertTrue(t.unit, u.s)

                # Check that we can compute flux dictionaries.
                f = model.get_initial_spectra(0*u.s, 10*u.MeV)
                self.assertEqual(type(f), dict)
                self.assertEqual(len(f), len(Flavor))
                self.assertEqual(f[Flavor.NU_E].unit, 1/(u.erg * u.s))

    def test_Walk_2019(self):
        """
        Instantiate a set of 'Walk 2019' models
        """

        for mass in [40.,75.]:
            for angles in [1,2]:
                model = Walk_2019(progenitor_mass=mass*u.Msun, direction=angles)

                self.assertEqual(model.metadata['EOS'], 'LS220')
                self.assertEqual(model.metadata['Progenitor mass'], mass*u.Msun)

                # Check that times are in proper units.
                t = model.get_time()
                self.assertTrue(t.unit, u.s)

                # Check that we can compute flux dictionaries.
                f = model.get_initial_spectra(0*u.s, 10*u.MeV)
                self.assertEqual(type(f), dict)
                self.assertEqual(len(f), len(Flavor))
                self.assertEqual(f[Flavor.NU_E].unit, 1/(u.erg * u.s))

    def test_Fornax_2019(self):
        """
        Instantiate a set of 'Fornax 2019' models
        """
        for mass in [9, 10, 12, 13, 14, 15, 19, 25, 60]:
            model = Fornax_2019(progenitor_mass=mass*u.Msun)

            self.assertEqual(model.metadata['Progenitor mass'].value, mass)

            # Check that times are in proper units.
            t = model.get_time()
            self.assertTrue(t.unit, u.s)

            # Check that we can compute flux dictionaries.
            f = model.get_initial_spectra(0*u.s, 10*u.MeV, theta=23*u.degree, phi=22*u.degree)
            self.assertEqual(type(f), dict)
            self.assertEqual(len(f), len(Flavor))
            self.assertEqual(f[Flavor.NU_E].unit, u.erg/(u.MeV * u.s))

    def test_Warren_2020(self):
        """
        Instantiate a set of 'Warren 2020' models
        """
        masses = np.concatenate((np.linspace(9.0, 12.75, 16),
                                 np.linspace(13, 30., 171),
                                 np.linspace(31., 33., 3),
                                 np.linspace(35, 55, 5),
                                 np.linspace(60, 80, 3),
                                 np.linspace(100, 120, 2)))

        for mixing in [1.23, 1.25, 1.27]:
            for mass in masses:
                model = Warren_2020(progenitor_mass=mass * u.Msun, turbmixing_param=mixing)

                self.assertEqual(model.metadata['Progenitor mass'], float(mass) * u.Msun)
                self.assertEqual(model.metadata['Turb. mixing param.'], mixing)
                self.assertEqual(model.metadata['EOS'], 'SFHo')

                # Check that times are in proper units.
                t = model.get_time()
                self.assertTrue(t.unit, u.s)

                # Check that we can compute flux dictionaries.
                f = model.get_initial_spectra(0 * u.s, 10 * u.MeV)
                self.assertEqual(type(f), dict)
                self.assertEqual(len(f), len(Flavor))
                self.assertEqual(f[Flavor.NU_E].unit, 1. / (u.erg * u.s))

    def test_Kuroda_2020(self):
        """
        Instantiate a set of 'Kuroda 2020' models
        """

        for rot_vel, exponents in [(0, [0]),
                                   (1, [12, 13])]:
            for exponent in exponents:
                model = Kuroda_2020(rotational_velocity=rot_vel * u.rad / u.s,
                                    magnetic_field_exponent=exponent, eos='LS220')

                self.assertEqual(model.metadata['EOS'], 'LS220')
                self.assertEqual(model.metadata['Progenitor mass'], 20*u.Msun)

                # Check that times are in proper units.
                t = model.get_time()
                self.assertTrue(t.unit, u.s)

                # Check that we can compute flux dictionaries.
                f = model.get_initial_spectra(0*u.s, 10*u.MeV)
                self.assertEqual(type(f), dict)
                self.assertEqual(len(f), len(Flavor))
                self.assertEqual(f[Flavor.NU_E].unit, 1/(u.erg * u.s))

    def test_Fornax_2021(self):
        """
        Instantiate a set of 'Fornax 2021' models
        """
        for mass in list(range(12, 24)) + [25, 26, 26.99]:
            model = Fornax_2021(progenitor_mass=mass*u.Msun)

            self.assertEqual(model.metadata['Progenitor mass'], float(mass)*u.Msun)

            # Check that times are in proper units.
            t = model.get_time()
            self.assertTrue(t.unit, u.s)

            # Check that we can compute flux dictionaries.
            f = model.get_initial_spectra(0*u.s, 10*u.MeV)
            self.assertEqual(type(f), dict)
            self.assertEqual(len(f), len(Flavor))
            self.assertEqual(f[Flavor.NU_E].unit, 1/(u.erg * u.s))

    def test_Fischer_2020(self):
        """
        Instantiate a set of 'Fischer 2020' models
        """
        model = Fischer_2020()
        self.assertEqual(model.metadata['Progenitor mass'], [18 * u.Msun])
        self.assertEqual(model.metadata['EOS'], 'HS(DD2)')

        # Check that times are in proper units.
        t = model.get_time()
        self.assertTrue(t.unit, u.s)

        # Check that we can compute flux dictionaries.
        f = model.get_initial_spectra(0*u.s, 10*u.MeV)
        self.assertEqual(type(f), dict)
        self.assertEqual(len(f), len(Flavor))
        self.assertEqual(f[Flavor.NU_E].unit, 1/(u.erg * u.s))

    def test_Zha_2021(self):
        """
        Instantiate a set of 'Zha 2021' models
        """

        for mass in list(range(16, 27)) + [19.89, 22.39, 30, 33]:
            model = Zha_2021(progenitor_mass=mass * u.Msun, eos='STOS_B145')

            self.assertEqual(model.metadata['Progenitor mass'], float(mass)*u.Msun)
            self.assertEqual(model.metadata['EOS'], 'STOS_B145')

            # Check that times are in proper units.
            t = model.get_time()
            self.assertTrue(t.unit, u.s)

            # Check that we can compute flux dictionaries.
            f = model.get_initial_spectra(0*u.s, 10*u.MeV)
            self.assertEqual(type(f), dict)
            self.assertEqual(len(f), len(Flavor))
            self.assertEqual(f[Flavor.NU_E].unit, 1./(u.erg * u.s))

    def test_Fornax_2022(self):
        """
        Instantiate a set of 'Fornax 2022' models
        """
        progenitors = Fornax_2022.param['progenitor_mass']
        bh_masses = [12.  , 12.03, 12.07, 12.1 , 12.18, 12.2 , 12.33, 12.4 , 12.45,
                      12.5 , 12.54, 12.6 , 12.72, 12.8 , 12.85, 12.9 , 12.97, 13.  ,
                      13.05, 13.25, 13.27, 13.32, 13.4 , 13.5 , 13.6 , 13.82, 13.9 ,
                      14.13, 14.25, 14.4 , 14.41, 14.44, 14.7 , 14.87, 15.  , 15.04,
                      15.38]<<u.Msun
        # PNS masses for each simulation, D. Vartanyan, private comm.
        pns_masses = [  1.35 , 1.38 , 1.4  , 1.45 , 1.5  ,
                        1.52 , 1.54 , 1.57 , 1.55 , 1.54 ,
                        1.55 , 1.57 , 1.69 , 1.73 , 1.72 ,
                        1.87 , 1.67 , 1.67 , 1.73 , 1.69 ,
                        1.68 , 1.73 , 1.75 , 1.75 , 1.73 ,
                        1.75 , 1.78 , 1.83 , 1.81 , 1.75 ,
                        1.84 , 1.74 , 1.8  , 1.74 , 1.86 ,
                        1.77 , 1.83 , 1.84 , 1.78 , 1.85 ,
                        1.79 , 1.87 , 1.87 , 1.84 , 1.75 ,
                        1.85 , 1.9  , 1.85 , 1.77 , 1.84 ,
                        1.87 , 1.79 , 1.77 , 1.87 , 1.82 ,
                        1.76 , 1.86 , 1.89 , 1.95 , 1.94 ,
                        1.69 , 1.94 , 1.74 , 1.99 , 1.97 ,
                        2.05 , 2.08 , 2.07 , 2.02 , 2.03 ,
                        1.72 , 1.68 , 2.12 , 1.81 , 1.83 ,
                        2.16 , 1.84 , 1.68 , 2.12 , 2.23 ,
                        1.85 , 2.39 , 2.34 , 2.31 , 1.99 ,
                        2.37 , 2.25 , 2.25 , 2.36 , 2.23 ,
                        2.1  , 2.16 , 1.98 , 2.01 , 2.1  ,
                        1.98 , 2.11 , 2.18 , 2.14 , 2.2 ]<<u.Msun

        for progenitor, m_pns in zip(progenitors, pns_masses):
            model = Fornax_2022(progenitor_mass=progenitor)

            self.assertEqual(model.metadata['Progenitor mass'], progenitor)
            self.assertEqual(model.metadata['PNS mass'], m_pns)
            self.assertEqual(model.metadata['Black hole'], progenitor in bh_masses)
            # Check that times are in proper units.
            t = model.get_time()
            self.assertTrue(t.unit, u.s)

            # Check that we can compute flux dictionaries.
            f = model.get_initial_spectra(0*u.s, 10*u.MeV)
            self.assertEqual(type(f), dict)
            self.assertEqual(len(f), len(Flavor))
            self.assertEqual(f[Flavor.NU_E].unit, 1/(u.erg * u.s))

    def test_Mori_2023(self):
        """
        Instantiate a set of 'Mori 2023' models
        """
        axion_mass_coupling = [ (0,0),
            (100,2), (100,4), (100,10), (100,12), (100,14), (100,16), (100,20),
            (200,2), (200,4), (200,6), (200,8), (200,10), (200,20)]

        pns_mass = [1.78, 1.77, 1.76, 1.77, 1.77, 1.77, 1.77, 1.74, 1.77, 1.76, 1.75, 1.74, 1.73, 1.62] * u.Msun

        for ((am, ac), mpns) in zip(axion_mass_coupling, pns_mass):
            model = Mori_2023(axion_mass=am*u.MeV, axion_coupling=ac*1e-10/u.GeV)

            axion_mass = float(am) * u.MeV
            self.assertEqual(model.metadata['Axion mass'], axion_mass)

            axion_coupling = float(ac) * 1e-10/u.GeV
            axion_coupling = np.round(axion_coupling.to('1e-10/GeV'))
            self.assertEqual(model.metadata['Axion coupling'], axion_coupling)

            self.assertEqual(model.metadata['Progenitor mass'], 20*u.Msun)
            self.assertEqual(model.metadata['PNS mass'], mpns)

            # Check that times are in proper units.
            t = model.get_time()
            self.assertTrue(t.unit, u.s)

            # Check that we can compute flux dictionaries.
            f = model.get_initial_spectra(0*u.s, 10*u.MeV)
            self.assertEqual(type(f), dict)
            self.assertEqual(len(f), len(Flavor))
            self.assertEqual(f[Flavor.NU_E].unit, 1./(u.erg * u.s))



import os
from tempfile import TemporaryDirectory

from hypothesis import given, assume, note
from hypothesis import strategies as st
from hypothesis.extra.numpy import arrays, array_shapes

from snewpy.flux import Container,Axes
from snewpy.neutrino import Flavor
import numpy as np
import astropy.units as u
import pytest

#define strategies to generate float values
def float_values(shape=array_shapes(max_dims=1),**kwargs):
    return arrays(float, shape=shape, 
                  elements={'allow_nan':False, 
                            'allow_infinity':False,
                            'min_value':-1e100,
                            'max_value':1e100
                           },
                  **kwargs)

def sorted_floats(shape=array_shapes(max_dims=1)):
    return float_values(shape, unique=True).map(np.sort)

def to_unit(unit):
    def _f(values):
        return values<<u.Unit(unit)

@st.composite
def quantities(draw, values, unit):
    return draw(values)<<unit

#how to generate axes values
flavors = st.lists(elements=st.sampled_from(Flavor),
                   unique=True,min_size=1,max_size=len(Flavor)).map(sorted)
units = st.sampled_from(['1/MeV','1/(MeV*s)','1/(MeV*s*m**2)']).map(u.Unit)
energies = quantities(sorted_floats(shape=array_shapes(max_dims=1)),u.MeV)
times    = quantities(sorted_floats(shape=array_shapes(max_dims=1)),u.s)
integrable_axess = st.lists(st.sampled_from(list(Axes)[1:]))

@st.composite
def random_flux_containers(draw, times=times, energies=energies, units=units, axes=st.just(None)):
    "Strategy to generate a random flux container"
    T,E = draw(times),draw(energies)
    flavor = draw(flavors)
    unit = draw(units)
    
    shape=[len(flavor),len(T),len(E)]
    array_data = quantities(float_values(shape),unit)
    array = draw(array_data)
    return Container[unit](array,flavor,T,E, integrable_axes=draw(axes))

#remove infinities and NaNs for the floats generation
st.register_type_strategy(float, st.floats(allow_nan=False, allow_infinity=False))

#starting test sets
#--------------------------------------------------------------
@given(flavor=flavors, time=times, energy=energies, unit=units)
def test_construction_succesfull(flavor, energy, time, unit):
    data = np.ones([len(flavor),len(time), len(energy)])<<unit
    assert Container[unit](data, flavor, time, energy)

@given(flavor=flavors, time=times, energy=energies, unit=units)
def test_construction_with_wrong_units_raises_ConversionError(flavor, energy, time, unit):
    data = np.ones([len(flavor),len(time), len(energy)])<<unit
    with pytest.raises(u.UnitConversionError):
        Container[unit*u.kg](data, flavor, time, energy)
    with pytest.raises(u.UnitConversionError):
        Container[unit](data*u.kg, flavor, time, energy)

@given(flavor=flavors, time=times, energy=energies, unit=units)
def test_construction_with_wrong_dimensions_raises_ValueError(flavor, energy, time, unit):
    data = np.ones([len(flavor)+1,len(time), len(energy)])<<unit
    with pytest.raises(ValueError):
        Container[unit](data, flavor, time, energy)

@given(f=random_flux_containers())
def test_summation_over_flavor(f:Container):
    fS = f.sum('flavor')
    #check the result array shapes
    assert fS.shape[Axes.flavor] == 1
    assert fS.shape[Axes.energy] == f.shape[Axes.energy]
    assert fS.shape[Axes.time]   == f.shape[Axes.time]
    #check the resulting array values
    assert np.allclose(fS.array, f.array.sum(axis=Axes.flavor))

axess_not_flavor = st.sampled_from([Axes.time,Axes.energy])

@given(f=random_flux_containers(), axis=axess_not_flavor
      )
def test_summation_over_wrong_axis_raises_ValueError(f:Container, axis:str):
    with pytest.raises(ValueError):
        fS = f.sum(axis)
    
@given(f=random_flux_containers())
def test_integration_over_flavor_raises_ValueError(f:Container):
    with pytest.raises(ValueError):
        fS = f.integrate('flavor')
    
@given(f=random_flux_containers(), axis=axess_not_flavor)
def test_integration_over_axes(f:Container, axis):
    #we don't want to integrate over single value
    assume(f.shape[axis]>1)
    note(f'Array is {str(f.array)}')
    fI = f.integrate(axis)
    for a in Axes:
        if a!=axis:
            assert fI.shape[a] == f.shape[a]
        else:
            assert fI.shape[a] == 1
    #check the resulting array values
    assert fI.unit == f.unit*f.axes[axis].unit

@given(f=random_flux_containers())
def test_save_and_load(f):
    with TemporaryDirectory() as tmpdir:
        fname = os.path.join(tmpdir, 'flux.npz')
        f.save(fname)
        f1 = Container.load(fname)
        assert f1 == f


import itertools as it
import os
from abc import ABC, abstractmethod
from warnings import warn

import numpy as np
from astropy import units as u
from astropy.table import Table
from astropy.units import UnitTypeError, get_physical_type
from astropy.units.quantity import Quantity
from scipy.special import loggamma
from snewpy._model_downloader import LocalFileLoader

from snewpy.neutrino import Flavor
from snewpy.flavor_transformation import NoTransformation
from functools import wraps

from snewpy.flux import Flux
from pathlib import Path

def _wrap_init(init, check):
    @wraps(init)
    def _wrapper(self, *arg, **kwargs):
        init(self, *arg, **kwargs)
        check(self)
    return _wrapper
    
class SupernovaModel(ABC, LocalFileLoader):
    """Base class defining an interface to a supernova model."""

    def __init_subclass__(cls, **kwargs):
        """Hook to modify the subclasses on creation"""
        super().__init_subclass__(**kwargs)
        cls.__init__ = _wrap_init(cls.__init__, cls.__post_init_check)

    def __init__(self, time, metadata):
        """Initialize supernova model base class
        (call this method in the subclass constructor as ``super().__init__(time,metadata)``).

        Parameters
        ----------
        time : ndarray of astropy.Quantity
            Time points where the model flux is defined.
            Must be array of :class:`Quantity`, with units convertable to "second".
        metadata : dict
            Dict of model parameters <name>:<value>,
            to be used for printing table in :meth:`__repr__` and :meth:`_repr_markdown_`
        """
        self.time = time
        self.metadata = metadata
        
    def __repr__(self):
        """Default representation of the model.
        """

        mod = f"{self.__class__.__name__} Model"
        try:
            mod += f': {self.filename}'
        except AttributeError:
            pass
        s = [mod]
        for name, v in self.metadata.items():
            s += [f"{name:16} : {v}"]
        return '\n'.join(s)

    def __post_init_check(self):
        """A function to check model integrity after initialization"""
        try:
            t = self.time
            m = self.metadata
        except AttributeError as e:
            clsname = self.__class__.__name__
            raise TypeError(f"Model not initialized. Please call 'SupernovaModel.__init__' within the '{clsname}.__init__'") from e

    def _repr_markdown_(self):
        """Markdown representation of the model, for Jupyter notebooks.
        """
        mod = f'**{self.__class__.__name__} Model**'
        try:
            mod +=f': {self.filename}'
        except:
            pass
        s = [mod,'']
        if self.metadata:
            s += ['|Parameter|Value|',
                  '|:--------|:----:|']
            for name, v in self.metadata.items():
                try: 
                    s += [f"|{name} | ${v.value:g}$ {v.unit:latex}|"]
                except:
                    s += [f"|{name} | {v} |"]
        return '\n'.join(s)

    def get_time(self):
        """Returns
        -------
        ndarray of astropy.Quantity
            Snapshot times from the simulation
        """
        return self.time

    @abstractmethod
    def get_initial_spectra(self, t, E, flavors=Flavor):
        """Get neutrino spectra at the source.

        Parameters
        ----------
        t : astropy.Quantity
            Time to evaluate initial spectra.
        E : astropy.Quantity or ndarray of astropy.Quantity
            Energies to evaluate the initial spectra.
        flavors: iterable of snewpy.neutrino.Flavor
            Return spectra for these flavors only (default: all)

        Returns
        -------
        initialspectra : dict
            Dictionary of neutrino spectra, keyed by neutrino flavor.
        """
        pass

    def get_initialspectra(self, *args):
        """DO NOT USE! Only for backward compatibility!

        :meta private:
        """
        warn("Please use `get_initial_spectra()` instead of `get_initialspectra()`!", FutureWarning)
        return self.get_initial_spectra(*args)

    def get_transformed_spectra(self, t, E, flavor_xform):
        """Get neutrino spectra after applying oscillation.

        Parameters
        ----------
        t : astropy.Quantity
            Time to evaluate initial and oscillated spectra.
        E : astropy.Quantity or ndarray of astropy.Quantity
            Energies to evaluate the initial and oscillated spectra.
        flavor_xform : FlavorTransformation
            An instance from the flavor_transformation module.

        Returns
        -------
        dict
            Dictionary of transformed spectra, keyed by neutrino flavor.
        """
        initialspectra = self.get_initial_spectra(t, E)
        transformed_spectra = {}

        transformed_spectra[Flavor.NU_E] = \
            flavor_xform.prob_ee(t, E) * initialspectra[Flavor.NU_E] + \
            flavor_xform.prob_ex(t, E) * initialspectra[Flavor.NU_X]

        transformed_spectra[Flavor.NU_X] = \
            flavor_xform.prob_xe(t, E) * initialspectra[Flavor.NU_E] + \
            flavor_xform.prob_xx(t, E) * initialspectra[Flavor.NU_X] 

        transformed_spectra[Flavor.NU_E_BAR] = \
            flavor_xform.prob_eebar(t, E) * initialspectra[Flavor.NU_E_BAR] + \
            flavor_xform.prob_exbar(t, E) * initialspectra[Flavor.NU_X_BAR]

        transformed_spectra[Flavor.NU_X_BAR] = \
            flavor_xform.prob_xebar(t, E) * initialspectra[Flavor.NU_E_BAR] + \
            flavor_xform.prob_xxbar(t, E) * initialspectra[Flavor.NU_X_BAR] 

        return transformed_spectra   

    def get_flux (self, t, E, distance, flavor_xform=NoTransformation()):
        """Get neutrino flux through 1cm^2 surface at the given distance

        Parameters
        ----------
        t : astropy.Quantity
            Time to evaluate the neutrino spectra.
        E : astropy.Quantity or ndarray of astropy.Quantity
            Energies to evaluate the the neutrino spectra.
        distance : astropy.Quantity or float (in kpc)
            Distance from supernova.
        flavor_xform : FlavorTransformation
            An instance from the flavor_transformation module.

        Returns
        -------
        dict
            Dictionary of neutrino fluxes in [neutrinos/(cm^2*erg*s)], 
            keyed by neutrino flavor.

        """
        distance = distance << u.kpc #assume that provided distance is in kpc, or convert
        factor = 1/(4*np.pi*(distance.to('cm'))**2)
        f = self.get_transformed_spectra(t, E, flavor_xform)

        array = np.stack([f[flv] for flv in sorted(Flavor)])
        return  Flux(data=array*factor, flavor=np.sort(Flavor), time=t, energy=E)



    def get_oscillatedspectra(self, *args):
        """DO NOT USE! Only for backward compatibility!

        :meta private:
        """
        warn("Please use `get_transformed_spectra()` instead of `get_oscillatedspectra()`!", FutureWarning)
        return self.get_transformed_spectra(*args)

def get_value(x):
    """If quantity x has is an astropy Quantity with units, return just the
    value.

    Parameters
    ----------
    x : Quantity, float, or ndarray
        Input quantity.

    Returns
    -------
    value : float or ndarray
    
    :meta private:
    """
    if type(x) == Quantity:
        return x.value
    return x

class PinchedModel(SupernovaModel):
    """Subclass that contains spectra/luminosity pinches"""
    def __init__(self, simtab, metadata):
        """ Initialize the PinchedModel using the data from the given table.

        Parameters
        ----------
        simtab: astropy.Table 
            Should contain columns TIME, {L,E,ALPHA}_NU_{E,E_BAR,X,X_BAR}
            The values for X_BAR may be missing, then NU_X data will be used
        metadata: dict
            Model parameters dict
        """
        if not 'L_NU_X_BAR' in simtab.colnames:
            # table only contains NU_E, NU_E_BAR, and NU_X, so double up
            # the use of NU_X for NU_X_BAR.
            for val in ['L','E','ALPHA']:
                simtab[f'{val}_NU_X_BAR'] = simtab[f'{val}_NU_X']
        # Get grid of model times.
        time = simtab['TIME'] << u.s
        # Set up dictionary of luminosity, mean energy and shape parameter
        # alpha, keyed by neutrino flavor (NU_E, NU_X, NU_E_BAR, NU_X_BAR).
        self.luminosity = {}
        self.meanE = {}
        self.pinch = {}
        for f in Flavor:
            self.luminosity[f] = simtab[f'L_{f.name}'] << u.erg/u.s
            self.meanE[f] = simtab[f'E_{f.name}'] << u.MeV
            self.pinch[f] = simtab[f'ALPHA_{f.name}']
        super().__init__(time, metadata)


    def get_initial_spectra(self, t, E, flavors=Flavor):
        """Get neutrino spectra/luminosity curves before oscillation.

        Parameters
        ----------
        t : astropy.Quantity
            Time to evaluate initial spectra.
        E : astropy.Quantity or ndarray of astropy.Quantity
            Energies to evaluate the initial spectra.
        flavors: iterable of snewpy.neutrino.Flavor
            Return spectra for these flavors only (default: all)

        Returns
        -------
        initialspectra : dict
            Dictionary of model spectra, keyed by neutrino flavor.
        """
        #convert input arguments to 1D arrays
        t = u.Quantity(t, ndmin=1)
        E = u.Quantity(E, ndmin=1)
        #Reshape the Energy array to shape [1,len(E)]
        E = np.expand_dims(E, axis=0)

        initialspectra = {}

        # Estimate L(t), <E_nu(t)> and alpha(t). Express all energies in erg.
        E = E.to_value('erg')

        # Make sure input time uses the same units as the model time grid, or
        # the interpolation will not work correctly.
        t = t.to(self.time.unit)

        for flavor in flavors:
            # Use np.interp rather than scipy.interpolate.interp1d because it
            # can handle dimensional units (astropy.Quantity).
            L  = get_value(np.interp(t, self.time, self.luminosity[flavor].to('erg/s')))
            Ea = get_value(np.interp(t, self.time, self.meanE[flavor].to('erg')))
            a  = np.interp(t, self.time, self.pinch[flavor])

            #Reshape the time-related arrays to shape [len(t),1]
            L  = np.expand_dims(L, axis=1)
            Ea = np.expand_dims(Ea,axis=1)
            a  = np.expand_dims(a, axis=1)

            # For numerical stability, evaluate log PDF and then exponentiate.
            # Suppress div-by-zero and other warnings that do not matter here.
            with np.errstate(divide='ignore', invalid='ignore'):
                result = \
                  np.exp(np.log(L) - (2+a)*np.log(Ea) + (1+a)*np.log(1+a)
                        - loggamma(1+a) + a*np.log(E) - (1+a)*(E/Ea)) / (u.erg * u.s)

            #remove bad values
            result[np.isnan(result)] = 0
            result[:, E[0]==0] = 0

            #remove unnecessary dimensions, if E or t was scalar:
            result = np.squeeze(result)
            initialspectra[flavor] = result

        return initialspectra



# -*- coding: utf-8 -*-
"""Utility module that reads a local YAML file listing the data files
available for the models supported by SNEWPY. These data files can either
be local (in the SNEWPY source tree) or uploaded to a model repository on
Zenodo. The YAML file supports regular expressions to allow matching of all
possible model files.
"""

import hashlib
import os
import requests
import yaml

from dataclasses import dataclass
from importlib.resources import open_text
from pathlib import Path
from tqdm.auto import tqdm
from typing import Optional

from snewpy import model_path
from snewpy import __version__ as snewpy_version

import logging

logger = logging.getLogger("FileHandle")


def _md5(fname: str) -> str:
    """calculate the md5sum hash of a file."""
    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def _download(src: str, dest: str, chunk_size=8192):
    """Download a file from 'src' to 'dest' and show the progress bar."""
    # make sure parent dir exists
    Path(dest).parent.mkdir(exist_ok=True, parents=True)
    with requests.get(src, stream=True) as r:
        r.raise_for_status()
        fileSize = int(r.headers.get("content-length", 0))
        with tqdm(
            desc=dest.name,
            total=fileSize,
            unit="iB",
            unit_scale=True,
            unit_divisor=1024,
        ) as bar:
            with open(dest, "wb") as f:
                for chunk in r.iter_content(chunk_size=chunk_size):
                    size = f.write(chunk)
                    bar.update(size)


class ChecksumError(FileNotFoundError):
    """Raise an exception due to a mismatch in the MD5 checksum."""

    def __init__(self, path, md5_exp, md5_actual):
        super().__init__(f"Checksum error for file {path}: {md5_actual}!={md5_exp}")

    pass


class MissingFileError(FileNotFoundError):
    """Raise an exception due to a missing file."""

    pass


@dataclass
class FileHandle:
    """Object storing local path, remote URL (optional), and MD5 sum
    (optional) for a SNEWPY model file. If the requested file is already present
    locally, open it. Otherwise, download it from the remote URL to the desired
    local path.
    """

    path: Path
    remote: str = None
    md5: Optional[str] = None

    def check(self) -> None:
        """Check if the given file exists locally and has a correct md5 sum.
        Raises
        ------
        :class:`MissingFileError`
            if the local copy of the file is missing
        :class:`ChecksumError`
            if the local file exists, but the checksum is wrong"""
        if not self.path.exists():
            raise MissingFileError(self.path)
        if self.md5:
            logger.info(f"File {self.path}: checking md5")
            md5 = _md5(self.path)
            logger.debug(f"{md5} vs expected {self.md5}")
            if md5 != self.md5:
                raise ChecksumError(self.path, self.md5, md5)

    def load(self) -> Path:
        """Make sure that local file exists and has a correct checksum.
        Download the file if needed.
        """
        try:
            self.check()
        except FileNotFoundError as e:
            logger.info(f"Downloading file {self.path}")
            _download(self.remote, self.path)
            self.check()
        return self.path


def _from_zenodo(zenodo_id: str, filename: str):
    """Access files on Zenodo.

    Parameters
    ----------
    zenodo_id : Zenodo record for model files.
    filename : Expected filename storing simulation data.

    Returns
    -------
    file_url, md5sum
    """
    zenodo_url = f"https://zenodo.org/api/records/{zenodo_id}"
    record = requests.get(zenodo_url).json()
    # Search for file string in Zenodo request for this record.
    file = next((_file for _file in record["files"] if _file["key"] == filename), None)

    # If matched, return a tuple of URL and checksum.Otherwise raise an exception.
    if file is not None:
        return file["links"]["self"], file["checksum"].split(":")[1]
    else:
        raise MissingFileError(filename)


class ModelRegistry:
    """Access model data. Configuration for each model is in a YAML file
    distributed with SNEWPY.
    """

    def __init__(self, config_file: Path = None, local_path: str = model_path):
        """
        Parameters
        ----------
        config_file: YAML configuration file. If None (default) use the 'model_files.yml' from SNEWPY resources
        local_path: local installation path (defaults to astropy cache).
        """
        if config_file is None:
            context = open_text("snewpy.models", "model_files.yml")
        else:
            context = open(config_file)
        with context as f:
            self.config = yaml.safe_load(f)
        self.models = self.config["models"]
        self.local_path = local_path

    def get_file(self, config_path: str, filename: str) -> Path:
        """Get the requested data file from the models file repository

        Parameters
        ----------
        config_path : dot-separated path of the model in the YAML configuration (e.g. "ccsn.Bollig_2016")
        filename : Name of simulation datafile, or a relative path from the model sub-directory

        Returns
        -------
        Path of downloaded file.
        """
        tokens = config_path.split(".")
        config = self.models
        for t in tokens:
            config = config[t]
        # store the model name
        model = tokens[-1]
        # Get data from GitHub or Zenodo.
        repo = config["repository"]
        if repo == "zenodo":
            url, md5 = _from_zenodo(zenodo_id=config["zenodo_id"], filename=filename)
        else:
            # format the url directly
            params = {
                "model": model,
                "filename": filename,
                "snewpy_version": snewpy_version,
            }
            params.update(
                config
            )  # default parameters can be overriden in the model config
            url, md5 = repo.format(**params), None
        localpath = Path(model_path) / str(model)
        localpath.mkdir(exist_ok=True, parents=True)
        fh = FileHandle(path=localpath / filename, remote=url, md5=md5)
        return fh.load()


registry = ModelRegistry()


class LocalFileLoader:
    @classmethod
    def request_file(cls, filename:str)->Path:
        "Require that the provided filename exists locally"
        path = Path(filename).absolute()
        if not path.exists():
            raise FileNotFoundError(path)
        return path

class RegistryFileLoader(LocalFileLoader):
    _registry = registry
    
    @classmethod
    def request_file(cls, filename:str)->Path:
        "Request file from the model registry"
        return cls._registry.get_file(cls._config_path, filename)
        
        
def get_model_data(model: str, filename: str) -> Path:
    """Get the requested data file from the models file repository

    Parameters
    ----------
    model : dot-separated path of the model in the YAML configuration (e.g. "ccsn.Bollig_2016")
    filename : Absolute path to simulation datafile, or a relative path from the model sub-directory

    Returns
    -------
    Path of downloaded file.
    """
    if os.path.isabs(filename):
        return Path(filename)
    else:
        return registry.get_file(model, filename)



# -*- coding: utf-8 -*-
"""
The submodule ``snewpy.models.ccsn_loaders`` contains classes to load core-collapse
supernova models from files stored on disk.
"""

import logging
import os
import re
import sys
import tarfile

from astropy import units as u
from astropy.table import Table, join
from astropy.io import ascii, fits
import h5py
import numpy as np
from scipy.special import gamma, lpmv

try:
    import healpy as hp
except ImportError:
    pass

from snewpy.models.base import PinchedModel, SupernovaModel
from snewpy.neutrino import Flavor
from snewpy import _model_downloader

class GarchingArchiveModel(PinchedModel):
    """Subclass that reads models in the format used in the
    `Garching Supernova Archive <https://wwwmpa.mpa-garching.mpg.de/ccsnarchive/>`_."""
    def __init__(self, filename, eos='LS220', metadata={}):
        """Model Initialization.

        Parameters
        ----------
        filename : str
            Absolute or relative path to file with model data, we add nue/nuebar/nux.  This argument will be deprecated.
        eos: str
            Equation of state. Valid value is 'LS220'. This argument will be deprecated.

        Other Parameters
        ----------------
        progenitor_mass: astropy.units.Quantity
            Mass of model progenitor in units Msun. Valid values are {progenitor_mass}.
        Raises
        ------
        FileNotFoundError
            If a file for the chosen model parameters cannot be found
        ValueError
            If a combination of parameters is invalid when loading from parameters
        """
        # Read through the several ASCII files for the chosen simulation and
        # merge the data into one giant table.
        mergtab = None
        for flavor in Flavor:
            _flav = Flavor.NU_X if flavor == Flavor.NU_X_BAR else flavor
            _sfx = _flav.name.replace('_', '').lower()
            _filename = '{}_{}_{}'.format(filename, eos, _sfx)
            _lname = 'L_{}'.format(flavor.name)
            _ename = 'E_{}'.format(flavor.name)
            _e2name = 'E2_{}'.format(flavor.name)
            _aname = 'ALPHA_{}'.format(flavor.name)

            # Open the requested filename using the model downloader.
            datafile = self.request_file(_filename)

            simtab = Table.read(datafile,
                                names=['TIME', _lname, _ename, _e2name],
                                format='ascii')
            simtab['TIME'].unit = 's'
            simtab[_lname].unit = '1e51 erg/s'
            simtab[_aname] = (2*simtab[_ename]**2 - simtab[_e2name]) / (simtab[_e2name] - simtab[_ename]**2)
            simtab[_ename].unit = 'MeV'
            del simtab[_e2name]

            if mergtab is None:
                mergtab = simtab
            else:
                mergtab = join(mergtab, simtab, keys='TIME', join_type='left')
                mergtab[_lname].fill_value = 0.
                mergtab[_ename].fill_value = 0.
                mergtab[_aname].fill_value = 0.
        simtab = mergtab.filled()
        if not metadata:
            metadata = {
                'Progenitor mass': float(os.path.basename(filename).split('s')[1].split('c')[0]) * u.Msun,
                'EOS': eos,
            }
        super().__init__(simtab, metadata)

class Nakazato_2013(PinchedModel):
    def __init__(self, filename, metadata={}):
        """Model initialization.

        Parameters
        ----------
        filename : str
            Absolute or relative path to FITS file with model data.

        Raises
        ------
        FileNotFoundError
            If a file for the chosen model parameters cannot be found
        """
        # Open the requested filename using the model downloader.
        datafile = self.request_file(filename)
        # Read FITS table using the astropy reader.
        simtab = Table.read(datafile)

        self.filename = os.path.basename(filename)
        super().__init__(simtab, metadata)


class Sukhbold_2015(Nakazato_2013):
    pass


class Tamborra_2014(GarchingArchiveModel):
    pass


class Bollig_2016(GarchingArchiveModel):
    pass


class Walk_2018(GarchingArchiveModel):
    pass


class Walk_2019(GarchingArchiveModel):
    pass


class OConnor_2013(PinchedModel):
    """Model based on the black hole formation simulation in `O'Connor & Ott (2013) <https://arxiv.org/abs/1207.1100>`_.
    """

    def __init__(self, filename, metadata={}):
        """
        Parameters
        ----------
        filename : str
            Absolute or relative path to FITS file with model data.
        """
        datafile = self.request_file(filename)
        # Open luminosity file.
        with tarfile.open(datafile) as tf:
            # Extract luminosity data.
            dataname = 's{:d}_{}_timeseries.dat'.format(int(metadata['Progenitor mass'].value), metadata['EOS'])
            # Read FITS table using the astropy reader.
            simtab = ascii.read(tf.extractfile(dataname), names=['TIME', 'L_NU_E', 'L_NU_E_BAR', 'L_NU_X',
                                                                    'E_NU_E', 'E_NU_E_BAR', 'E_NU_X',
                                                                    'RMS_NU_E', 'RMS_NU_E_BAR', 'RMS_NU_X'])

        simtab['ALPHA_NU_E'] = (2.0 * simtab['E_NU_E'] ** 2 - simtab['RMS_NU_E'] ** 2) / (
                simtab['RMS_NU_E'] ** 2 - simtab['E_NU_E'] ** 2)
        simtab['ALPHA_NU_E_BAR'] = (2.0 * simtab['E_NU_E_BAR'] ** 2 - simtab['RMS_NU_E_BAR'] ** 2) / (
                simtab['RMS_NU_E_BAR'] ** 2 - simtab['E_NU_E_BAR'] ** 2)
        simtab['ALPHA_NU_X'] = (2.0 * simtab['E_NU_X'] ** 2 - simtab['RMS_NU_X'] ** 2) / (
                simtab['RMS_NU_X'] ** 2 - simtab['E_NU_X'] ** 2)

        # note, here L_NU_X is already divided by 4
        super().__init__(simtab, metadata)


class OConnor_2015(PinchedModel):
    """Model based on the black hole formation simulation in `O'Connor (2015) <https://arxiv.org/abs/1411.7058>`_.
    """

    def __init__(self, filename, metadata={}):
        """
        Parameters
        ----------
        filename : str
            Absolute or relative path to FITS file with model data.
        """

        datafile = self.request_file(filename)
        simtab = Table.read(datafile,
                            names=['TIME', 'L_NU_E', 'L_NU_E_BAR', 'L_NU_X',
                                    'E_NU_E', 'E_NU_E_BAR', 'E_NU_X',
                                    'RMS_NU_E', 'RMS_NU_E_BAR', 'RMS_NU_X'],
                            format='ascii')

        header = ascii.read(simtab.meta['comments'], delimiter='=', format='no_header', names=['key', 'val'])
        tbounce = float(header['val'][0])
        simtab['TIME'] -= tbounce

        simtab['ALPHA_NU_E'] = (2.0*simtab['E_NU_E']**2 - simtab['RMS_NU_E']**2) / \
            (simtab['RMS_NU_E']**2 - simtab['E_NU_E']**2)
        simtab['ALPHA_NU_E_BAR'] = (2.0*simtab['E_NU_E_BAR']**2 - simtab['RMS_NU_E_BAR']**2) / \
            (simtab['RMS_NU_E_BAR']**2 - simtab['E_NU_E_BAR']**2)
        simtab['ALPHA_NU_X'] = (2.0*simtab['E_NU_X']**2 - simtab['RMS_NU_X']**2) / \
            (simtab['RMS_NU_X']**2 - simtab['E_NU_X']**2)

        # SYB: double-check on this factor of 4. Should be factor of 2?
        simtab['L_NU_X'] /= 4.0

        # prevent negative lums
        simtab['L_NU_E'][simtab['L_NU_E'] < 0] = 1
        simtab['L_NU_E_BAR'][simtab['L_NU_E_BAR'] < 0] = 1
        simtab['L_NU_X'][simtab['L_NU_X'] < 0] = 1

        self.filename = os.path.basename(filename)

        super().__init__(simtab, metadata)


class Zha_2021(OConnor_2015):
    pass

class Warren_2020(PinchedModel):
    def __init__(self, filename, metadata={}):
        """
        Parameters
        ----------
        filename : str
            Absolute or relative path to file prefix, we add nue/nuebar/nux
        """
        # Open the requested filename using the model downloader.
        datafile = self.request_file(filename)

        # Open luminosity file.
        # Read data from HDF5 files, then store.
        f = h5py.File(datafile, 'r')

        simtab = Table()

        for i in range(len(f['nue_data']['lum'])):
            if f['sim_data']['shock_radius'][i][1] > 0.00001:
                bounce = f['sim_data']['shock_radius'][i][0]
                break

        simtab['TIME'] = f['nue_data']['lum'][:, 0] - bounce
        simtab['L_NU_E'] = f['nue_data']['lum'][:, 1] * 1e51
        simtab['L_NU_E_BAR'] = f['nuae_data']['lum'][:, 1] * 1e51
        simtab['L_NU_X'] = f['nux_data']['lum'][:, 1] * 1e51
        simtab['E_NU_E'] = f['nue_data']['avg_energy'][:, 1]
        simtab['E_NU_E_BAR'] = f['nuae_data']['avg_energy'][:, 1]
        simtab['E_NU_X'] = f['nux_data']['avg_energy'][:, 1]
        simtab['RMS_NU_E'] = f['nue_data']['rms_energy'][:, 1]
        simtab['RMS_NU_E_BAR'] = f['nuae_data']['rms_energy'][:, 1]
        simtab['RMS_NU_X'] = f['nux_data']['rms_energy'][:, 1]

        simtab['ALPHA_NU_E'] = (2.0 * simtab['E_NU_E'] ** 2 - simtab['RMS_NU_E'] ** 2) / \
            (simtab['RMS_NU_E'] ** 2 - simtab['E_NU_E'] ** 2)
        simtab['ALPHA_NU_E_BAR'] = (2.0 * simtab['E_NU_E_BAR'] ** 2 - simtab['RMS_NU_E_BAR']
                                    ** 2) / (simtab['RMS_NU_E_BAR'] ** 2 - simtab['E_NU_E_BAR'] ** 2)
        simtab['ALPHA_NU_X'] = (2.0 * simtab['E_NU_X'] ** 2 - simtab['RMS_NU_X'] ** 2) / \
            (simtab['RMS_NU_X'] ** 2 - simtab['E_NU_X'] ** 2)

        # Set model metadata.
        self.filename = os.path.basename(filename)

        super().__init__(simtab, metadata)


class Kuroda_2020(PinchedModel):
    def __init__(self, filename, metadata={}):
        """
        Parameters
        ----------
        filename : str
            Absolute or relative path to file prefix, we add nue/nuebar/nux
        """

        # Open the requested filename using the model downloader.
        datafile = self.request_file(filename)
        # Read ASCII data.
        simtab = Table.read(datafile, format='ascii')

        # Get grid of model times.
        simtab['TIME'] = simtab['Tpb[ms]'] << u.ms
        for f in [Flavor.NU_E, Flavor.NU_E_BAR, Flavor.NU_X]:
            fkey = re.sub('(E|X)_BAR', r'A\g<1>', f.name).lower()
            simtab[f'L_{f.name}'] = simtab[f'<L{fkey}>'] * 1e51 << u.erg / u.s
            simtab[f'E_{f.name}'] = simtab[f'<E{fkey}>'] << u.MeV
            # There is no pinch parameter so use alpha=2.0.
            simtab[f'ALPHA_{f.name}'] = np.full_like(simtab[f'E_{f.name}'].value, 2.)

        self.filename = os.path.basename(filename)

        super().__init__(simtab, metadata)


class Fornax_2019(SupernovaModel):
    def __init__(self, filename, metadata={}, cache_flux=False):
        """
        Parameters
        ----------
        filename : str
            Absolute or relative path to FITS file with model data.
        cache_flux : bool
            If true, pre-compute the flux on a fixed angular grid and store the values in a FITS file.
        """
        # Set up model metadata.
        self.filename = filename
        self.metadata = metadata

        self.fluxunit = 1e50 * u.erg/(u.s*u.MeV)
        self.time = None

        # Read a cached flux file in FITS format or generate one.
        self.is_cached = cache_flux and 'healpy' in sys.modules
        if cache_flux and not 'healpy' in sys.modules:
            logger = logging.getLogger()
            logger.warning("No module named 'healpy'. Cannot enable caching.")

        if self.is_cached:

            self.E = {}
            self.dE = {}
            self.dLdE = {}
            self.luminosity = {}

            # Check if we're initializing on a FITS file or not.
            if filename.endswith('.fits'):
                fitsfile = filename
            else:
                fitsfile = filename.replace('h5', 'fits')

            if os.path.exists(fitsfile):
                self._read_fits(fitsfile)
                ntim, nene, npix = self.dLdE[Flavor.NU_E].shape
                self.npix = npix
                self.nside = hp.npix2nside(npix)
            else:
                with h5py.File(filename, 'r') as _h5file:
                    # Conversion of flavor to key name in the model HDF5 file.
                    self._flavorkeys = {Flavor.NU_E: 'nu0',
                                        Flavor.NU_E_BAR: 'nu1',
                                        Flavor.NU_X: 'nu2',
                                        Flavor.NU_X_BAR: 'nu2'}

                    if self.time is None:
                        self.time = _h5file['nu0']['g0'].attrs['time'] * u.s

                    # Use a HEALPix grid with nside=4 (192 pixels) to cache the
                    # values of Y_lm(theta, phi).
                    self.nside = 4
                    self.npix = hp.nside2npix(self.nside)
                    thetac, phic = hp.pix2ang(self.nside, np.arange(self.npix))

                    Ylm = {}
                    for l in range(3):
                        Ylm[l] = {}
                        for m in range(-l, l+1):
                            Ylm[l][m] = self._real_sph_harm(l, m, thetac, phic)

                    # Store 3D tables of dL/dE for each flavor.
                    logger = logging.getLogger()
                    for flavor in Flavor:

                        key = self._flavorkeys[flavor]
                        logger.info('Caching {} for {} ({})'.format(filename, str(flavor), key))

                        # HDF5 file only contains NU_E, NU_E_BAR, and NU_X.
                        if flavor == Flavor.NU_X_BAR:
                            self.E[flavor] = self.E[Flavor.NU_X]
                            self.dE[flavor] = self.dE[Flavor.NU_X]
                            self.dLdE[flavor] = self.dLdE[Flavor.NU_X]
                            self.luminosity[flavor] = self.luminosity[Flavor.NU_X]
                            continue

                        self.E[flavor] = _h5file[key]['egroup'][()] * u.MeV
                        self.dE[flavor] = _h5file[key]['degroup'][()] * u.MeV

                        ntim, nene = self.E[flavor].shape
                        self.dLdE[flavor] = np.zeros((ntim, nene, self.npix), dtype=float)
                        # Loop over time bins.
                        for i in range(ntim):
                            # Loop over energy bins.
                            for j in range(nene):
                                dLdE_ij = 0.
                                # Sum over multipole moments.
                                for l in range(3):
                                    for m in range(-l, l+1):
                                        dLdE_ij += _h5file[key]['g{}'.format(j)
                                                                ]['l={} m={}'.format(l, m)][i] * Ylm[l][m]
                                self.dLdE[flavor][i][j] = dLdE_ij

                        # Integrate over energy to get L(t).
                        factor = 1. if flavor.is_electron else 0.25
                        self.dLdE[flavor] = self.dLdE[flavor] * factor * self.fluxunit
                        self.dLdE[flavor] = self.dLdE[flavor].to('erg/(s*MeV)')

                        self.luminosity[flavor] = np.sum(self.dLdE[flavor] * self.dE[flavor][:, :, np.newaxis], axis=1)

                    # Write output to FITS.
                    self._write_fits(fitsfile, overwrite=True)
        else:
            # Conversion of flavor to key name in the model HDF5 file.
            self._flavorkeys = {Flavor.NU_E: 'nu0',
                                Flavor.NU_E_BAR: 'nu1',
                                Flavor.NU_X: 'nu2',
                                Flavor.NU_X_BAR: 'nu2'}

            # Open the requested filename using the model downloader.
            datafile = self.request_file(filename)
            # Open HDF5 data file.
            self._h5file = h5py.File(datafile, 'r')

            # Get grid of model times in seconds.
            self.time = self._h5file['nu0']['g0'].attrs['time'] * u.s

    def _read_fits(self, filename):
        """Read cached angular data from FITS.

        Parameters
        ----------
        filename : str
            Input filename.
        """
        hdus = fits.open(filename)

        self.time = hdus['TIME'].data * u.Unit(hdus['TIME'].header['BUNIT'])

        for flavor in Flavor:
            name = str(flavor).split('.')[-1]

            ext = '{}_ENERGY'.format(name)
            self.E[flavor] = hdus[ext].data * u.Unit(hdus[ext].header['BUNIT'])

            ext = '{}_DE'.format(name)
            self.dE[flavor] = hdus[ext].data * u.Unit(hdus[ext].header['BUNIT'])

            ext = '{}_FLUX'.format(name)
            self.dLdE[flavor] = hdus[ext].data * u.Unit(hdus[ext].header['BUNIT'])
            self.dLdE[flavor] = self.dLdE[flavor].to('erg/(s*MeV)')

            self.luminosity[flavor] = np.sum(self.dLdE[flavor] * self.dE[flavor][:, :, np.newaxis], axis=1)

    def _write_fits(self, filename, overwrite=False):
        """Write angular-dependent calculated flux in FITS format.

        Parameters
        ----------
        filename : str
            Output filename.
        """
        hx = fits.HDUList()

        hdu_time = fits.PrimaryHDU(self.time.to_value('s'))
        hdu_time.header['EXTNAME'] = 'TIME'
        hdu_time.header['BUNIT'] = 'second'
        hx.append(hdu_time)

        for flavor in Flavor:
            name = str(flavor).split('.')[-1]

            hdu_E = fits.ImageHDU(self.E[flavor].to_value('MeV'))
            hdu_E.header['EXTNAME'] = '{}_ENERGY'.format(name)
            hdu_E.header['BUNIT'] = 'MeV'
            hx.append(hdu_E)

            hdu_dE = fits.ImageHDU(self.dE[flavor].to_value('MeV'))
            hdu_dE.header['EXTNAME'] = '{}_DE'.format(name)
            hdu_dE.header['BUNIT'] = 'MeV'
            hx.append(hdu_dE)

            hdu_flux = fits.ImageHDU(self.dLdE[flavor].to_value(str(self.fluxunit)))
            hdu_flux.header['EXTNAME'] = '{}_FLUX'.format(name)
            hdu_flux.header['BUNIT'] = str(self.fluxunit)
            hx.append(hdu_flux)

        hx.writeto(filename, overwrite=overwrite)

    def _fact(self, n):
        """Calculate n!.

        Parameters
        ----------
        n : int or float
            Input for computing n factorial.

        Returns
        -------
        factorial : float
            Factorial n!, computed as Gamma(n+1).
        """
        return gamma(n + 1.)

    def _real_sph_harm(self, l, m, theta, phi):
        """Compute orthonormalized real (tesseral) spherical harmonics Y_lm.

        Parameters
        ----------
        l : int
            Degree of the spherical harmonics.
        m : int
            Order of the spherical harmonics.
        theta : float or ndarray
            Input zenith angles.
        phi : float or ndarray
            Input azimuth angles.

        Returns
        -------
        Y_lm : float or ndarray
            Real-valued spherical harmonic function at theta, phi.
        """
        if m < 0:
            norm = np.sqrt((2*l + 1.)/(2*np.pi)*self._fact(l + m)/self._fact(l - m))
            return norm * lpmv(-m, l, np.cos(theta)) * np.sin(-m*phi)
        elif m == 0:
            norm = np.sqrt((2*l + 1.)/(4*np.pi))
            return norm * lpmv(0, l, np.cos(theta)) * np.ones_like(phi)
        else:
            norm = np.sqrt((2*l + 1.)/(2*np.pi)*self._fact(l - m)/self._fact(l + m))
            return norm * lpmv(m, l, np.cos(theta)) * np.cos(m*phi)

    def _get_binnedspectra(self, t, theta, phi):
        """Get binned neutrino spectrum at a particular time.

        Parameters
        ----------
        t : float or astropy.Quantity
            Time to evaluate initial and oscillated spectra.
        theta : astropy.Quantity
            Zenith angle of the spectral emission.
        phi : astropy.Quantity
            Azimuth angle of the spectral emission.

        Returns
        -------
        E : dict
            Dictionary of energy bin central values, keyed by neutrino flavor.
        dE : dict
            Dictionary of energy bin widths, keyed by neutrino flavor.
        binspec : dict
            Dictionary of binned model spectra, keyed by neutrino flavor.
        """
        E = {}
        dE = {}
        binspec = {}

        # Convert input time to a time index.
        t = t.to(self.time.unit)
        j = (np.abs(t - self.time)).argmin()

        for flavor in Flavor:
            # Cached data: read out the relevant time and angular rows.
            if self.is_cached:
                # Convert input angles to a HEALPix index.
                k = hp.ang2pix(self.nside, theta.to_value('radian'), phi.to_value('radian'))
                E[flavor] = self.E[flavor][j]
                dE[flavor] = self.dE[flavor][j]
                binspec[flavor] = self.dLdE[flavor][j, :, k]

            # Read the HDF5 input file directly and extract the spectra.
            else:
                # File only contains NU_E, NU_E_BAR, and NU_X.
                if flavor == Flavor.NU_X_BAR:
                    E[flavor] = E[Flavor.NU_X]
                    dE[flavor] = dE[Flavor.NU_X]
                    binspec[flavor] = binspec[Flavor.NU_X]
                    continue

                key = self._flavorkeys[flavor]

                # Energy binning of the model for this flavor, in units of MeV.
                E[flavor] = self._h5file[key]['egroup'][j] * u.MeV
                dE[flavor] = self._h5file[key]['degroup'][j] * u.MeV

                # Storage of differential flux per energy, angle, and time.
                dLdE = np.zeros(len(E[flavor]), dtype=float)

                # Loop over energy bins.
                for ebin in range(len(E[flavor])):
                    dLdE_j = 0
                    # Sum over multipole moments.
                    for l in range(3):
                        for m in range(-l, l + 1):
                            Ylm = self._real_sph_harm(l, m, theta.to_value('radian'), phi.to_value('radian'))
                            dLdE_j += self._h5file[key]['g{}'.format(ebin)]['l={} m={}'.format(l, m)][j] * Ylm
                    dLdE[ebin] = dLdE_j

                factor = 1. if flavor.is_electron else 0.25
                binspec[flavor] = dLdE * factor * self.fluxunit
                binspec[flavor] = binspec[flavor].to('erg/(s*MeV)')

        return E, dE, binspec

    def get_initial_spectra(self, t, E, theta, phi, flavors=Flavor, interpolation='linear'):
        """Get neutrino spectra/luminosity curves before flavor transformation.

        Parameters
        ----------
        t : astropy.Quantity
            Time to evaluate initial spectra.
        E : astropy.Quantity or ndarray of astropy.Quantity
            Energies to evaluate the initial spectra.
        theta : astropy.Quantity
            Zenith angle of the spectral emission.
        phi : astropy.Quantity
            Azimuth angle of the spectral emission.
        flavors: iterable of snewpy.neutrino.Flavor
            Return spectra for these flavors only (default: all)
        interpolation : str
            Scheme to interpolate in spectra ('nearest', 'linear').

        Returns
        -------
        initialspectra : dict
            Dictionary of model spectra, keyed by neutrino flavor.
        """
        initialspectra = {}

        # Extract the binned spectra for the input t, theta, phi:
        _E, _dE, _spec = self._get_binnedspectra(t, theta, phi)

        # Avoid "division by zero" in retrieval of the spectrum.
        E[E == 0] = np.finfo(float).eps * E.unit
        logE = np.log10(E.to_value('MeV'))

        for flavor in flavors:

            # Linear interpolation in flux.
            if interpolation.lower() == 'linear':
                # Pad log(E) array with values where flux is fixed to zero.
                _logE = np.log10(_E[flavor].to_value('MeV'))
                _dlogE = np.diff(_logE)
                _logEbins = np.insert(_logE, 0, np.log10(np.finfo(float).eps * E.unit/u.MeV))
                _logEbins = np.append(_logEbins, _logE[-1] + _dlogE[-1])

                # Pad with values where flux is fixed to zero.
                _dLdE = _spec[flavor].to_value(self.fluxunit)
                _dLdE = np.insert(_dLdE, 0, 0.)
                _dLdE = np.append(_dLdE, 0.)

                initialspectra[flavor] = np.interp(logE, _logEbins, _dLdE) * self.fluxunit

            elif interpolation.lower() == 'nearest':
                _logE = np.log10(_E[flavor].to_value('MeV'))
                _dlogE = np.diff(_logE)[0]
                _logEbins = _logE - _dlogE
                _logEbins = np.concatenate((_logEbins, [_logE[-1] + _dlogE]))
                _Ebins = 10**_logEbins

                idx = np.searchsorted(_Ebins, E) - 1
                select = (idx > 0) & (idx < len(_E[flavor]))

                _dLdE = np.zeros(len(E))
                _dLdE[np.where(select)] = np.asarray([_spec[flavor][i].to_value(self.fluxunit) for i in idx[select]])
                initialspectra[flavor] = _dLdE * self.fluxunit

            else:
                raise ValueError('Unrecognized interpolation type "{}"'.format(interpolation))

        return initialspectra


class Fornax_2021(SupernovaModel):
    def __init__(self, filename, metadata={}):
        """
        Parameters
        ----------
        filename : str
            Absolute or relative path to HDF5 file with model data.
        """

        # Open the requested filename using the model downloader.
        datafile = self.request_file(filename)
        # Set up model metadata.
        self.progenitor_mass = float(filename.split('/')[-1].split('_')[2][:-1]) * u.Msun
        self.metadata = metadata
        # Open HDF5 data file.
        _h5file = h5py.File(datafile, 'r')

        self.time = _h5file['nu0'].attrs['time'] * u.s

        self.luminosity = {}
        self._E = {}
        self._dLdE = {}
        for flavor in Flavor:
            # Convert flavor to key name in the model HDF5 file
            key = {Flavor.NU_E: 'nu0',
                   Flavor.NU_E_BAR: 'nu1',
                   Flavor.NU_X: 'nu2',
                   Flavor.NU_X_BAR: 'nu2'}[flavor]

            self._E[flavor] = np.asarray(_h5file[key]['egroup'])
            self._dLdE[flavor] = {f"g{i}": np.asarray(_h5file[key][f'g{i}']) for i in range(12)}

            # Compute luminosity by integrating over model energy bins.
            dE = np.asarray(_h5file[key]['degroup'])
            n = len(dE[0])
            dLdE = np.zeros((len(self.time), n), dtype=float)
            for i in range(n):
                dLdE[:, i] = self._dLdE[flavor][f"g{i}"]

            # Note factor of 0.25 in nu_x and nu_x_bar.
            factor = 1. if flavor.is_electron else 0.25
            self.luminosity[flavor] = np.sum(dLdE*dE, axis=1) * factor * 1e50 * u.erg/u.s

    def get_initial_spectra(self, t, E, flavors=Flavor, interpolation='linear'):
        """Get neutrino spectra/luminosity curves after oscillation.

        Parameters
        ----------
        t : astropy.Quantity
            Time to evaluate initial spectra.
        E : astropy.Quantity or ndarray of astropy.Quantity
            Energies to evaluate the initial spectra.
        flavors: iterable of snewpy.neutrino.Flavor
            Return spectra for these flavors only (default: all)
        interpolation : str
            Scheme to interpolate in spectra ('nearest', 'linear').

        Returns
        -------
        initialspectra : dict
            Dictionary of model spectra, keyed by neutrino flavor.
        """
        initialspectra = {}

        # Avoid "division by zero" in retrieval of the spectrum.
        E[E == 0] = np.finfo(float).eps * E.unit
        logE = np.log10(E.to_value('MeV'))

        # Make sure the input time uses the same units as the model time grid.
        # Convert input time to a time index.
        t = u.Quantity(t.to(self.time.unit), ndmin=1)
        j = np.array(list(np.abs(_t - self.time).argmin() for _t in t))

        for flavor in flavors:
            # Energy bin centers (in MeV)
            _E = self._E[flavor][j]
            _logE = np.log10(_E)
            _dlogE = np.diff(_logE)

            # Model flavors (internally) are nu_e, nu_e_bar, and nu_x, which stands
            # for nu_mu(_bar) and nu_tau(_bar), making the flux 4x higher than nu_e and nu_e_bar.
            factor = 1. if flavor.is_electron else 0.25

            # Linear interpolation in flux.
            if interpolation.lower() == 'linear':
                # Pad log(E) array with values where flux is fixed to zero.
                _logEbins = np.insert(_logE, 0, np.log10(np.finfo(float).eps * E.unit/u.MeV), axis=1)
                _logEbins = np.append(_logEbins, np.expand_dims(_logE[:,-1] + _dlogE[:,-1], 1), axis=1)

                # Luminosity spectrum _dLdE is in units of 1e50 erg/s/MeV.
                # Pad with values where flux is fixed to zero, then divide by E to get number luminosity
                _dNLdE = np.asarray([np.zeros(j.shape)] + [self._dLdE[flavor]['g{}'.format(i)][j] for i in range(12)] + [np.zeros(j.shape)]).T
                interp_values = np.array([np.interp(logE, __logEbins, __dNLdE)
                                          for __logEbins, __dNLdE in zip(_logEbins, _dNLdE)])
                initialspectra[flavor] = (interp_values / E * factor * 1e50 * u.erg/u.s/u.MeV).to('1 / (erg s)')

            elif interpolation.lower() == 'nearest':
                # Find edges of energy bins and identify which energy bin (each entry of) E falls into
                _logEbinEdges = _logE - _dlogE[0,0] / 2
                _logEbinEdges = np.append(_logEbinEdges, np.expand_dims(_logE[:,-1] + _dlogE[:,-1]/2, 1), axis=1)
                _EbinEdges = 10**_logEbinEdges
                idx = np.array([np.searchsorted(edges, E) - 1 for edges in _EbinEdges])
                select = np.array([(_idx > 0) & (_idx < len(__E)) for _idx, __E in zip(idx, _E)])

                # Divide luminosity spectrum by energy at bin center to get number luminosity spectrum
                _dNLdE = np.zeros([len(j), len(np.atleast_1d(E))])
                for i in range(len(j)):
                    _dNLdE[i][np.where(select[i])] = np.asarray([self._dLdE[flavor]['g{}'.format(ebin_idx)][j[i]] / _E[i][ebin_idx]
                                                                 for ebin_idx in idx[i][select[i]]])
                initialspectra[flavor] = ((_dNLdE << 1/u.MeV) * factor * 1e50 * u.erg/u.s/u.MeV).to('1 / (erg s)')

            else:
                raise ValueError('Unrecognized interpolation type "{}"'.format(interpolation))

        return initialspectra


class Fornax_2022(Fornax_2021):
    def __init__(self, filename, metadata={}):
        """
        Parameters
        ----------
        filename : str
            Absolute or relative path to HDF5 file with model data.
        """
        # Open the requested filename using the model downloader.
        datafile = self.request_file(filename)
        # Set up model metadata.
        self.progenitor = os.path.splitext(os.path.basename(filename))[0].split('_')[2]
        self.progenitor_mass = float(self.progenitor[:-3])*u.Msun if self.progenitor.endswith('bh') else float(self.progenitor)*u.Msun

        self.metadata = metadata

        # Open HDF5 data file.
        _h5file = h5py.File(datafile, 'r')

        self.metadata['PNS mass'] = _h5file.attrs['Mpns'] * u.Msun
        self.time = _h5file['nu0'].attrs['time'] * u.s

        self.luminosity = {}
        self._E = {}
        self._dLdE = {}
        for flavor in Flavor:
            # Convert flavor to key name in the model HDF5 file
            key = {Flavor.NU_E: 'nu0',
                   Flavor.NU_E_BAR: 'nu1',
                   Flavor.NU_X: 'nu2',
                   Flavor.NU_X_BAR: 'nu2'}[flavor]

            self._E[flavor] = np.asarray(_h5file[key]['egroup'])
            self._dLdE[flavor] = {f"g{i}": np.asarray(_h5file[key][f'g{i}']) for i in range(12)}

            # Compute luminosity by integrating over model energy bins.
            dE = np.asarray(_h5file[key]['degroup'])
            n = len(dE[0])
            dLdE = np.zeros((len(self.time), n), dtype=float)
            for i in range(n):
                dLdE[:, i] = self._dLdE[flavor][f"g{i}"]

            # Note factor of 0.25 in nu_x and nu_x_bar.
            factor = 1. if flavor.is_electron else 0.25
            self.luminosity[flavor] = np.sum(dLdE*dE, axis=1) * factor * 1e50 * u.erg/u.s


class Mori_2023(PinchedModel):
    def __init__(self, filename, metadata={}):
        """
        Parameters
        ----------
        filename : str
            Absolute or relative path to file prefix.
        """
        # Open the requested filename using the model downloader.
        datafile = self.request_file(filename)

        self.metadata = metadata

        # Read ASCII data.
        simtab = Table.read(datafile, format='ascii')

        # Remove the first table row, which appears to have zero input.
        simtab = simtab[simtab['1:t_sim[s]'] > 0]

        # Get grid of model times.
        simtab['TIME'] = simtab['2:t_pb[s]'] << u.s
        for j, (f, fkey) in enumerate(zip([Flavor.NU_E, Flavor.NU_E_BAR, Flavor.NU_X], 'ebx')):
            simtab[f'L_{f.name}'] = simtab[f'{6+j}:Le{fkey}[e/s]'] << u.erg / u.s
            # Compute the pinch parameter from E_rms and E_avg
            # <E^2> / <E>^2 = (2+a)/(1+a), where
            # E_rms^2 = <E^2> - <E>^2.
            Eavg = simtab[f'{9+j}:Em{fkey}[MeV]']
            Erms = simtab[f'{12+j}:Er{fkey}[MeV]']
            E2 = Erms**2 + Eavg**2
            x = E2 / Eavg**2
            alpha = (2-x) / (x-1)

            simtab[f'E_{f.name}'] = Eavg << u.MeV
            simtab[f'E2_{f.name}'] = E2 << u.MeV**2
            simtab[f'ALPHA_{f.name}'] = alpha

#            simtab[f'E_{f.name}'] = simtab[f'{9+j}:Em{fkey}[MeV]'] << u.MeV
#            Erms = simtab[f'{12+j}:Er{fkey}[MeV]'] * u.MeV
#
#            # Compute the pinch parameter from E_rms and E_avg
#            simtab[f'E2_{f.name}'] = Erms**2 + simtab[f'E_{f.name}']**2
#            x = simtab[f'E2_{f.name}'] / simtab[f'E_{f.name}']**2
#            simtab[f'ALPHA_{f.name}'] = (2-x) / (x-1)

        self.filename = os.path.basename(filename)

        super().__init__(simtab, metadata)


class Bugli_2021(PinchedModel):
    """Model based on `Buggli (2021) <https://arxiv.org/abs/2105.00665>`_.
    """

    def __init__(self, filename, metadata={}):
        """
        Parameters
        ----------
        filename : str
            Absolute or relative path to FITS file with model data.
        """

        datafile = self.request_file(filename)
        simtab = Table.read(datafile,
                            names=['TIME', 'L_NU_E', 'L_NU_E_BAR', 'L_NU_X',
                                    'E_NU_E', 'E_NU_E_BAR', 'E_NU_X',
                                    'RMS_NU_E', 'RMS_NU_E_BAR', 'RMS_NU_X'],
                            format='ascii')

        simtab['ALPHA_NU_E'] = (2.0*simtab['E_NU_E']**2 - simtab['RMS_NU_E']**2) / \
            (simtab['RMS_NU_E']**2 - simtab['E_NU_E']**2)
        simtab['ALPHA_NU_E_BAR'] = (2.0*simtab['E_NU_E_BAR']**2 - simtab['RMS_NU_E_BAR']**2) / \
            (simtab['RMS_NU_E_BAR']**2 - simtab['E_NU_E_BAR']**2)
        simtab['ALPHA_NU_X'] = (2.0*simtab['E_NU_X']**2 - simtab['RMS_NU_X']**2) / \
            (simtab['RMS_NU_X']**2 - simtab['E_NU_X']**2)

        self.filename = os.path.basename(filename)

        super().__init__(simtab, metadata)


class Fischer_2020(PinchedModel):
    def __init__(self, filename, metadata={}):
        """
        Parameters
        ----------
        filename : str
            Absolute or relative path to file
        """
        # Open the requested filename using the model downloader.
        datafile = self.request_file(filename)
        self.metadata = metadata

        # Open the requested filename using the model downloader.
        # datafile = _model_downloader.get_model_data(self.__class__.__name__, filename)
        # self.filename = os.path.basename(filename)

        simtab = Table()

        tf = tarfile.open(datafile)

        # Open luminosity file
        with tf.extractfile("luminosity.dat") as Lfile:
             Ldata = np.genfromtxt(Lfile, skip_header=2)

             simtab['TIME'] = Ldata[:, 0]

             simtab['L_NU_E'] = Ldata[:, 1]
             simtab['L_NU_E_BAR'] = Ldata[:, 2]
             simtab['L_NU_X'] = Ldata[:, 3]
             simtab['L_NU_X_BAR'] = Ldata[:, 4]

             Lfile.close()

        # Open mean energy file
        with tf.extractfile("menergy.dat") as Efile:
            Edata = np.genfromtxt(Efile, skip_header=2)

            simtab['E_NU_E'] = Edata[:, 1] << u.MeV
            simtab['E_NU_E_BAR'] = Edata[:, 2] << u.MeV
            simtab['E_NU_X'] = Edata[:, 3] << u.MeV
            simtab['E_NU_X_BAR'] = Edata[:, 4] << u.MeV

            Efile.close()

        # Open rms energy file
        with tf.extractfile("rmsenergy.dat") as RMSEfile:
            RMSEdata = np.genfromtxt(RMSEfile, skip_header=2)

            simtab['RMS_NU_E'] = RMSEdata[:, 1] << u.MeV
            simtab['RMS_NU_E_BAR'] = RMSEdata[:, 2] << u.MeV
            simtab['RMS_NU_X'] = RMSEdata[:, 3] << u.MeV
            simtab['RMS_NU_X_BAR'] = RMSEdata[:, 4] << u.MeV

            RMSEfile.close()

        simtab['ALPHA_NU_E'] = (2.0 * simtab['E_NU_E'] ** 2 - simtab['RMS_NU_E'] ** 2) / \
            (simtab['RMS_NU_E'] ** 2 - simtab['E_NU_E'] ** 2)
        simtab['ALPHA_NU_E_BAR'] = (2.0 * simtab['E_NU_E_BAR'] ** 2 - simtab['RMS_NU_E_BAR'] ** 2) / \
            (simtab['RMS_NU_E_BAR'] ** 2 - simtab['E_NU_E_BAR'] ** 2)
        simtab['ALPHA_NU_X'] = (2.0 * simtab['E_NU_X'] ** 2 - simtab['RMS_NU_X'] ** 2) / \
            (simtab['RMS_NU_X'] ** 2 - simtab['E_NU_X'] ** 2)
        simtab['ALPHA_NU_X_BAR'] = (2.0 * simtab['E_NU_X_BAR'] ** 2 - simtab['RMS_NU_X_BAR'] ** 2) / \
            (simtab['RMS_NU_X_BAR'] ** 2 - simtab['E_NU_X_BAR'] ** 2)

        tf.close()

        super().__init__(simtab, metadata)



"""
This module contains utilities to define a new registry supernova model.

Examples usage can be seen in :mod:`snewpy.models.ccsn`.


.. autoclass:: Parameter
    :members:
    
.. autodecorator:: RegistryModel
"""
from functools import wraps
import itertools as it
import inspect
import os
from textwrap import dedent
from warnings import warn
import numpy as np
import re
from snewpy._model_downloader import RegistryFileLoader

from astropy import table
from tqdm.auto import tqdm

all_models = set()
    
def get_models_table(init:bool=False)->table.QTable:
    """Get the astropy.Table with all the possible Metadata parameters.

    Parameters
    ----------
    init:bool
        If true - initialize each model to obtain the runtime metadata parameters (takes longer)
        Default is false, i.e. use only input parameters
    """
    #prepare the table
    tables = {}
    for model in tqdm(all_models):
        param_combinations = model.get_param_combinations()
        if(init):
            metadata = [model(**params).metadata for params in param_combinations]
        else:
            metadata = [{model.parameters[key].label:val for key,val in params.items()} for params in param_combinations]
            
        md_table = table.QTable(metadata)
        md_table['model']='.'.join([model.__module__,model.__name__])
        md_table['init_params']=param_combinations
        #set the fixed metadata
        if hasattr(model,"__metadata__"):
            for key,value in model.__metadata__.items():
                md_table[key]=value
        tables[model]=md_table
    
    df = table.vstack(list(tables.values()))
    return df
    
def _can_decorate_class_or_func(func_decorator):
    """make the function decorator act as class decorator:
    if decorated object is a class, wrap its "__init__" function.
    """
    def _wrapper(obj):
        if inspect.isclass(obj):
            obj.__init__=func_decorator(obj.__init__)
            return obj
        else:
            return func_decorator(obj)
    return _wrapper
            
def deprecated(*names, message='Argument `{name}` is deprecated.'):
    """A function decorator to issue a deprecation warning if a given argument is provided in the wrapped function call.
    
    Parameters
    ----------
    
    names: list of str
        argument names which are deprecated
    message: str
        a template with {name} parameter, to make the message for each argument.

    .. Example::

    @deprecated('foo','bar', message='Argument `{name}` is deprecated and will be removed in SNEWPYv2.0!')
    def test_function(*, foo=1, bar=2, baz=3):
        pass
        
    #calling test_function(foo=1, baz=3) will issue a deprecation warning:
    #    DeprecationWarning: Argument `foo` is deprecated and will be removed in SNEWPYv2.0!
    """
    @_can_decorate_class_or_func
    def _wrapper(func):
        #get function signature
        S = inspect.signature(func)
        @wraps(func)
        def _f(*args, **kwargs):
            #bind signature to find all the parameters
            params = S.bind(*args,**kwargs)
            for name in names:
                if name in params.arguments:
                    warn(message.format(name=name), category=UserWarning, stacklevel=2)
            return func(*args,**kwargs)
        return _f
    return _wrapper

def map_arguments(**names_dict):
    """map function arguments to create a function with new signature,
    and with updated docstring
    
    Parameters
    ----------
    names_dict: dict
        Argument names in format {old_name:new_name}
    
    Examples
    --------
    >>> @map_arguments(bar='BAR')
    >>> def foo(bar, baz):
    >>>     print(f'bar={bar} baz={baz}')
    >>>
    >>> foo(BAR=123, baz=10) 
    >>> #bar=123 baz=10
    """
    def _wrapper(func):
        S = inspect.signature(func) #old signature
        old2new = names_dict
        new2old = {v:k for k,v in old2new.items()}
        
        @wraps(func)
        def func_new(*args, **kwargs):
            params = S1.bind(*args,**kwargs)
            kwargs_old = params.kwargs
            kwargs_new = {new2old.get(name,name):val for name,val in kwargs_old.items()}
            return func(*params.args,**kwargs_new)
            
        #update signature
        S1 = S.replace(parameters=[
            p.replace(name=old2new.get(p.name,p.name)) for p in S.parameters.values()
            ])
        func_new.__signature__=S1
        #update the docstring
        def _update_docs(docs):
            for name_old,name_new in old2new.items():
                docs = re.sub(rf'\n{name_old}([\s:].*\n)',rf'\n{name_new}\g<1>',docs)
            return docs
        func_new.__doc__ = _update_docs(func.__doc__)
        if(hasattr(func_new,'__init__')):
            func_new.__init__.__doc__ = _update_docs(func.__init__.__doc__)
        return func_new
    return _wrapper
    
def _expand_defaults(func, **defaults):
    """Update the signature of the given function with the parameters with default values.
    Examples
    --------
    >>> def foo(x, y=1):
    >>>     pass
    >>> foo?
    Signature: foo(x, y=1)
    >>> _expand_defaults(foo, x=123, z=456)
    >>> foo?
    Signature: foo(x=123, y=500, *, z=456)
    """
    S = inspect.signature(func)
    params = list(S.parameters.values())
    #defaults.update(**{p.name:p.default for p in params if p.default!=p.empty})
    params_updated = [p.replace(default=defaults.pop(p.name, p.default)) for p in params]
    #add the remaining default parameters
    params_updated+=[inspect.Parameter(name,kind=inspect.Parameter.KEYWORD_ONLY, default=value) for name,value in defaults.items()]
    func.__signature__ = S.replace(parameters=params_updated)

_default_labels={'eos':'EOS', 'magnetic_field_exponent':'B_0 Exponent'}
_default_descriptions={'eos':'Equation of state',
                       'progenitor_mass':'Mass of model progenitor in units Msun',
                       'revival_time':'Time of shock revival in model in units ms',
                       'metallicity':'Progenitor metallicity',
                       'turbmixing_param':'Turbulent mixing parameter alpha_lambda',
                       'rotational_velocity': 'Rotational velocity of progenitor',
                       'magnetic_field_exponent':'Exponent of magnetic field'
                      }
class Parameter:
    """A class to describe the model parameter: it's range of allowed values, name, description etc. """
    def __init__(self, values, *,
                 name:str=None, 
                 label:str=None, 
                 description:str=None, 
                 desc_values:str=None,
                 precision:int=None):
        """
        Parameters
        ----------
        values : iterable
            List of allowed parameter values. If len(values)==1 the parameter is considered fixed (this can be used to add default metadata to the model).
            
        Other Parameters
        ------------------
        name : str
            a variable name for the parameter (used to set the ``label`` and ``description`` if they are `None`)
        label : str
            A key to be used for this parameter in the metadata dictionary (e.g. 'EOS' for name='eos')
            If `None`(default), label will be derived from name
        description : str
            A long description of the parameter, to be used in the docstring generation
            If `None`(default), description will be derived from name
        desc_values:str
            A string representation of the given values range, to be used in the docstring generation
            If `None`(default), desc_values will be just `str(values)`
        precision:int or None
            Number of decimals, used to round the input value if when testing if given element is in the list of allowed values. This should be used only if the `values` are an array of floats or `Quantity`.
            If `None` (default) no rounding is applied.
            
        If label or description is `None` they are derived from the name, 
        by capitalizing and replacing underscores with spaces.
        
        Examples
        --------
        
        >>> # create a parameter object:
        >>> Parameter(range(0,100,1),name='supernova_parameter', desc_values='[0,1,...99]')
        Parameter(name="supernova_parameter", label="Supernova parameter", description="Supernova parameter", values='[0,1,...99]')
        """
        self.values = values
        self.label = label 
        self.description = description
        self.desc_values = desc_values or str(values)
        self.precision = precision
        if(name):
            self.set_description_and_label(name)
        
    def set_description_and_label(self, name:str):
        """Generate the description and label from the parameter given name"""
        self.label = self.label or _default_labels.get(name, name.replace('_',' ').capitalize())
        self.description = self.description or _default_descriptions.get(name,self.label)
        
    def apply_precision(self, value):
        if self.precision:
            if hasattr(self.values, 'unit'):
                #convert unit before 
                value = value<<self.values.unit 
            value = np.around(value,decimals=self.precision)
        return value
        
    def __contains__(self, value):
        value = self.apply_precision(value)
        return value in self.values
    def __iter__(self):
        return self.values.__iter__()
    def __len__(self):
        return self.values.__len__()
    def __getitem__(self, index):
        return self.values.__getitem__(index)
    def __repr__(self):
        return f"{self.__class__.__name__}(label=\"{self.label}\", description=\"{self.description}\", values={self.desc_values})"
    @property
    def fixed(self):
        """True if this parameter has only one option"""
        return len(self)==1


class ParameterSet:
    """A class to describe all possible combinations of parameter values"""
    def __init__(self, param_validator=None, **params):
        """
        Parameters
        ----------
        params:dict
            Keyword arguments, passing each parameter as iterable or instance of :class:`Parameter`
        param_validator:callable or None
            A function of user parameters (dict), returning true if the passed user parameters are valid.
            If ``None`` (default) - all the combinations are allowed
        """
        #construct Parameter classes from given arrays
        self.params = {}
        for name,val in params.items():
            if not isinstance(val,Parameter):
                val = Parameter(values=val,name=name)
            else:
                val.set_description_and_label(name)
            self.params[name]=val
        #fill the valid parameter combinations
        self.combinations = list(it.product(*[list(v) for v in self.params.values()]))
        self.valid_combinations = []
        if param_validator:
            for p in self.combinations:
                pdict = dict(zip(params.keys(),p))
                is_valid = param_validator(pdict)
                if is_valid:
                    self.valid_combinations+=[p]
        else:
            self.valid_combinations = self.combinations
        self.valid_combinations_dict = [dict(zip(self.params,p)) for p in self.valid_combinations]
        #fill the default values for fixed parameters
        self.defaults = {name:p[0] for name,p in self.params.items() if p.fixed}
            
    def _fill_default_parameters(self, **user_params)->dict:
        """Set the parameter values if they are missing in 'user_params' and fixed (i.e. have a single option"""
        return dict(self.defaults, **user_params)

    def _apply_precision(self, **user_params)->dict:
        """Check that we have all required parameters in `user_params`.
        Apply parameter precision and unit conversion for every given parameter"""
        for name,val in user_params.items():
            if name in self.params:
                user_params[name]=self.params[name].apply_precision(val)
        return user_params
        
    def _check_all_parameters_present(self, **user_params):
        """Check that `user_params` contains all needed parameters, and nothing extra"""
        #check that we have all correct parameters
        for name in user_params:
            if name not in self.params:
                raise ValueError(f"Unexpected parameter '{name}', allowed parameters are {list(self.params.keys())}")
        for name in self.params:
            if name not in user_params:
                raise ValueError(f"Missing parameter '{name}'")
                
    def _check_parameter_values(self, **user_params):
        """Check that provided user parameters are valid, i.e. are within the given list values"""
        for name in user_params:
            if user_params[name] not in self.params[name]:
                raise ValueError(f"Invalid parameter value for '{name}'={user_params[name]}. Allowed values are: {self.params[name].desc_values}")
        #check that parameter combination is valid
        ptuple = tuple([user_params[name] for name in self.params])
        if not ptuple in self.valid_combinations:
            valid_dicts = [dict(zip(self.params.keys(),pars)) for pars in self.valid_combinations]
            raise ValueError(f"Invalid parameters combination: {user_params}")
            
    def validate(self, **user_params):
        """Check that provided user parameters are valid, i.e. are within the given list values"""
        #check that we have all correct parameters
        self._check_all_parameters_present(**user_params)
        user_params = self._apply_precision(**user_params)
        self._check_parameter_values(**user_params)

    def __getitem__(self, name:str):
        return self.params.__getitem__(name)
    def items(self):
        return self.params.items()
    def values(self):
        return self.params.values()
    def __iter__(self):
        return self.params.__iter__()
    def __repr__(self):
        s = f"{self.__class__.__name__}:\n"
        s+='\n'.join([f' * \t{name}={ps}' for name,ps in self.params.items()])
        return s

    def generate_docstring(self, func, **type_annotations)->str:
        #generate docstring
        s = []
        S = inspect.signature(func)
        params = [name for name in self.params.keys() if name in S.parameters]
        for name in params:
            p = self.params[name]
            p_type = type_annotations.get(name,None)
            type_name = p_type.__name__ if p_type else ''
            s+=[f'{name}: {type_name}\n    {p.description}. Valid values are: {p.desc_values}.']
        return '\n'.join(s)
    
def RegistryModel(_param_validator=None, **params):
    """A class decorator for defining the supernova model, that initializes from physics parameters.
    
    Parameters
    -----------
    params:dict
        Keyword arguments, passing each parameter as iterable or instance of :class:`Parameter`
    _param_validator:callable or None
        A function of user parameters (dict), returning true if the passed user parameters are valid.
        If `None` (default) - all the combinations are allowed

    Returns
    -------
        Model class
            A class, inherited from the input class, which has modified `__init__` function (see details below)
    
    Note
    ----
    This decorator helps to construct the daughter of the given class, which automatically implements the constructor with:
    
        * Validation of the input user parameters (see :meth:`RegistryModel.validate`)
        * Generated constructor docstring based on allowed parameters
        * Populates the `self.metadata` from the given user parameters
        * Optional (deprecated) "filename" argument, and calls initialization from the filename

        The decorated class:
            * *must* inherit from the loader class
            * *must* implement the __init__ method, which
            
                1. defines the filename from the given user parameters
                2. optionally modify the `self.metadata` dictionary, and 
                3. calls the corresponding loader class constructor
            * *can* define the ``_metadata_from_filename (self, filename:str)->dict``, to help populate the metadata when a `filename` argument is passed

    If a parameter has a single allowed value (i.e. fixed), this value will be default for this parameter in constructor.
    
    """
    pset:ParameterSet = ParameterSet(param_validator=_param_validator, **params)
    def _wrap(base_class):
        class c(base_class, RegistryFileLoader):
            parameters:ParameterSet = pset
            @classmethod
            def _generate_docstring(cls)->str:
                docstring = dedent(base_class.__init__.__doc__ or '')
                for section, desc in cls._doc_params_.items():
                    s = f'{section}\n'+'-'*len(section)+'\n'+dedent(desc)
                    docstring+='\n'+s+'\n'
                return docstring
                
            def __init__(self, *args, **kwargs):
                # enforce the default parameters
                params = inspect.signature(self.__init__).bind(*args, **kwargs)
                params.apply_defaults()
                #select the parameters which correspond to metadata
                arguments = {name:val for name,val in params.arguments.items() if name in self.parameters}
                arguments = self.parameters._fill_default_parameters(**arguments)
                arguments = self.parameters._apply_precision(**arguments)
                # validate the input parameters
                self.parameters.validate(**arguments)
                #Store model metadata
                self.metadata = {self.parameters[name].label: value for name,value in arguments.items()}
                #call the constructor with only the needed arguments (the rest are only for metadata)
                S = inspect.signature(super().__init__)
                init_params = {name:val for name,val in arguments.items() if name in S.parameters}
                return super().__init__(**init_params)
                
            @classmethod
            def get_param_combinations(cls)->tuple:
                """Get all valid combinations of parameters for a this registry model.
                Returns
                -------
                valid_combinations: tuple[dict]
                    A tuple of all valid parameter combinations stored as Dictionaries"""
                return cls.parameters.valid_combinations_dict
            
            param = {name: p.values for name, p in parameters.items()}
        #fill the constructor signature
        c.__init__.__signature__ = inspect.signature(base_class.__init__)
        #If we have "fixed" parameters in ParameterSet, add them to the signature as keyword arguments
        defaults = c.parameters.defaults
        _expand_defaults(c.__init__, **defaults)
        #generate the docstring
        c._doc_params_ = {
                'Other parameters': pset.generate_docstring(c.__init__, **c.__init__.__annotations__),
                'Raises':"""
                FileNotFoundError
                    If a file for the chosen model parameters cannot be found
                ValueError
                    If a combination of parameters is invalid when loading from parameters"""
            }
        c.__doc__ = base_class.__doc__
        c.__init__.__doc__ = c._generate_docstring()
        #fill the class and module name to be the same as in class
        c.__qualname__ = base_class.__qualname__
        c.__name__ = base_class.__name__
        c.__module__ = base_class.__module__
        #set the configuration path
        if not hasattr(c, '_config_path'):
            module_name = c.__module__.split(".")[-1]
            c._config_path = f'{module_name}.{c.__name__}'
        #register the model in the list
        global all_models
        all_models.add(c)
        return c
    return _wrap

def legacy_filename_initialization(c):
    """Wrap the model class, adding a filename argument in the init"""
    
    @deprecated('filename')
    class c1(c):
        _loader_class = c.__mro__[2] #store the loader class for later use

        def __init__(self, filename:str=None, *args, **kwargs):
            if filename is not None:
                if not hasattr(self,'metadata'):
                    self.metadata = {}
                if hasattr(self,'_metadata_from_filename'):
                    self.metadata.update(self._metadata_from_filename(filename))
                self._loader_class.__init__(self, filename=os.path.abspath(filename), metadata=self.metadata)
            else:
                super().__init__(*args, **kwargs)

    #generate the docstring
    c1.__doc__ = c.__doc__
    c1._doc_params_ = {'Parameters':
                       """filename: str\n    Absolute or relative path to the file with model data. This argument is deprecated.""",
                       **c._doc_params_}
    c1.__init__.__doc__ = c1._generate_docstring()
    #update the call signature
    S = inspect.signature(c)
    S1 = inspect.signature(c1.__init__)
    #set default values to None if they are not set
    other_params = []
    for p in S.parameters.values():
        if p.default==p.empty:
            p = p.replace(default=None)
        other_params+=[p]
    params = [S1.parameters['self'],S1.parameters['filename'],*other_params]
    #fill the constructor signature
    c1.__init__.__signature__ = S.replace(parameters=params)
    #fill the class and module name to be the same as in class
    c1.__qualname__ = c.__qualname__
    c1.__name__ = c.__name__
    c1.__module__ = c.__module__
    #register the model in the list
    global all_models
    all_models.remove(c)
    all_models.add(c1)
    return c1



# -*- coding: utf-8 -*-
"""
The submodule ``snewpy.models.ccsn`` contains models of core-collapse supernovae
derived from the :class:`SupernovaModel` base class.

Since SNEWPY v1.3, the prefered method is to initialise a model based on its physics parameters.
Initialisation from a file name is deprecated.

Use the ``param`` class property to view all physics parameters and their possible values:

>>> from snewpy.models.ccsn import Nakazato_2013
>>> Nakazato_2013.param
{'progenitor_mass': <Quantity [13., 20., 30., 50.] solMass>,
 'revival_time': <Quantity [  0., 100., 200., 300.] ms>,
 'metallicity': [0.02, 0.004],
 'eos': ['LS220', 'shen', 'togashi']}

For some models, not all combinations of parameters are valid. Use the ``get_param_combinations()``
class method to get a list of all valid combinations and filter it:

>>> list(params for params in Nakazato_2013.get_param_combinations() if params['eos'] != 'shen')
[{'progenitor_mass': <Quantity 30. solMass>, 'revival_time': <Quantity 0. ms>,
  'metallicity': 0.004, 'eos': 'LS220'},
 {'progenitor_mass': <Quantity 30. solMass>, 'revival_time': <Quantity 0. ms>,
  'metallicity': 0.004, 'eos': 'togashi'}]

.. _Garching Supernova Archive: https://wwwmpa.mpa-garching.mpg.de/ccsnarchive/

"""
import logging
import os
import re
import tarfile

import numpy as np
from astropy import units as u
from astropy.table import Table

from snewpy.models import ccsn_loaders as loaders
from .base import PinchedModel

from snewpy.models.registry_model import RegistryModel, Parameter
from snewpy.models.registry_model import deprecated, legacy_filename_initialization
from snewpy.models.registry_model import all_models
from textwrap import dedent


@RegistryModel()
class Fischer_2020(loaders.Fischer_2020):
    """Model based on simulations from `Fischer et al. (2020) <https://arxiv.org/abs/1804.10890>`
    """
    def __init__(self):
        self.metadata["EOS"] = "HS(DD2)"
        self.metadata["Progenitor mass"] = 18 * u.Msun
        filename='Fischer_2020.tar.gz'
        return super().__init__(filename, metadata=self.metadata)

@legacy_filename_initialization
@RegistryModel(
    progenitor_mass = [13, 20, 30, 50] * u.Msun,
    revival_time = [0, 100, 200, 300] * u.ms,
    metallicity = [0.02, 0.004],
    eos = ['LS220', 'shen', 'togashi'],

    _param_validator = lambda p: (p['revival_time'] == 0 * u.ms and p['progenitor_mass'] == 30 * u.Msun
                                  and p['metallicity'] == 0.004) or \
                                 (p['revival_time'] != 0 * u.ms and p['eos'] == 'shen'
                                  and not (p['progenitor_mass'] == 30 * u.Msun and p['metallicity'] == 0.004))
)
class Nakazato_2013(loaders.Nakazato_2013):
    """Model based on simulations from Nakazato et al., ApJ S 205:2
    (2013), ApJ 804:75 (2015), PASJ 73:639 (2021). See also http://asphwww.ph.noda.tus.ac.jp/snn/.
    """

    def _metadata_from_filename(self, filename:str)->dict:
            metadata = {'Progenitor mass': float(filename.split('-')[-1].strip('s%.fits')) * u.Msun}
            if 't_rev' in filename:
                metadata.update({
                    'EOS': filename.split('-')[-4].upper(),
                    'Metallicity': float(filename.split('-')[-3].strip('z%')),
                    'Revival time': float(filename.split('-')[-2].strip('t_rev%ms')) * u.ms
                })
            # No revival time because the explosion "failed" (BH formation).
            else:
                metadata.update({
                    'EOS': filename.split('-')[-4].upper(),
                    'Metallicity': float(filename.split('-')[-2].strip('z%')),
                    'Revival time': 0 * u.ms
                })
            return metadata

    def __init__(self, progenitor_mass:u.Quantity, revival_time:u.Quantity, metallicity:float, eos:str):
        # Strip units for filename construction
        progenitor_mass = progenitor_mass.to(u.Msun).value
        revival_time = revival_time.to(u.ms).value

        if revival_time != 0:
            filename = f"nakazato-{eos}-z{metallicity}-t_rev{int(revival_time)}ms-s{progenitor_mass:3.1f}.fits"
        else:
            filename = f"nakazato-{eos}-BH-z{metallicity}-s{progenitor_mass:3.1f}.fits"
        #modify metadata if needed...
        #self.metadata['name']=value
        return super().__init__(filename, self.metadata)


@legacy_filename_initialization
@RegistryModel(
    progenitor_mass = [27., 9.6] * u.Msun,
    eos = ['LS220', 'SFHo']
)
class Sukhbold_2015(loaders.Sukhbold_2015):
    """Model based on simulations from Sukhbold et al., ApJ 821:38,2016. Models were shared privately by email.
    """
    def _metadata_from_filename(self, filename:str):
        metadata = {
            'Progenitor mass': float(filename.split('-')[-1].strip('z%.fits')) * u.Msun,
            'EOS': filename.split('-')[-2]
        }
        return metadata

    def __init__(self, progenitor_mass:u.Quantity, eos:str):
        if progenitor_mass.value == 9.6:
            filename = f'sukhbold-{eos}-z{progenitor_mass.value:3.1f}.fits'
        else:
            filename = f'sukhbold-{eos}-s{progenitor_mass.value:3.1f}.fits'
        return super().__init__(filename, self.metadata)


@deprecated('eos')
@legacy_filename_initialization
@RegistryModel(
    progenitor_mass = [11.2, 20., 27.] * u.Msun,
    direction = [1,2,3],
    eos=['LS220'],
    _param_validator = lambda p: (p['progenitor_mass'] == 11.2 * u.Msun and p['direction'] in (1,)) or \
        (p['progenitor_mass'] == 20. * u.Msun and p['direction'] in (1,)) or \
        (p['progenitor_mass'] == 27. * u.Msun and p['direction'] in (1,2,3))
)
class Tamborra_2014(loaders.Tamborra_2014):
    """Model based on 3D simulations from `Tamborra et al., PRD 90:045032, 2014 <https://arxiv.org/abs/1406.0006>`_.
    Data files are from the `Garching Supernova Archive`_.
    """
    def __init__(self, *, progenitor_mass:u.Quantity, direction:int):
        filename = f's{progenitor_mass.value:3.1f}c_3D_dir{direction}'
        # Metadata is handled by __init__ in _GarchingArchiveModel
        return super().__init__(filename=filename, metadata=self.metadata)


@deprecated('eos')
@legacy_filename_initialization
@RegistryModel(
    progenitor_mass= [11.2, 27.] * u.Msun,
    eos = ['LS220']
)
class Bollig_2016(loaders.Bollig_2016):
    """Model based on simulations from `Bollig et al. (2016) <https://arxiv.org/abs/1508.00785>`_.
    Models were taken, with permission, from the `Garching Supernova Archive`_.
    """
    def __init__(self, progenitor_mass:u.Quantity, eos:str='LS220'):
        filename = f's{progenitor_mass.value:3.1f}c'
        return super().__init__(filename=filename, metadata=self.metadata)


@deprecated('eos')
@legacy_filename_initialization
@RegistryModel(
    progenitor_mass = [15.] * u.Msun,
    rotation = ['fast','slow','non'],
    direction = [1,2,3],
    eos=['LS220']
)
class Walk_2018(loaders.Walk_2018):
    """Model based on SASI-dominated simulations from `Walk et al.,
    PRD 98:123001, 2018 <https://arxiv.org/abs/1807.02366>`_. Data files are from
    the `Garching Supernova Archive`_.
    """
    def __init__(self, *, progenitor_mass:u.Quantity, rotation:str, direction:int):

        filename = f's{progenitor_mass.value:3.1f}c_3D_{rotation}rot_dir{direction}'
        return super().__init__(filename=filename, metadata=self.metadata)


@deprecated('eos')
@legacy_filename_initialization
@RegistryModel(
    progenitor_mass = [40., 75.] * u.Msun,
    direction = [1,2,3],
    eos=['LS220'],
    _param_validator = lambda p: (p['progenitor_mass'] == 75. * u.Msun and p['direction'] in (1,2)) or \
        (p['progenitor_mass'] == 40. * u.Msun and p['direction'] in (1,2,3))
)
class Walk_2019(loaders.Walk_2019):
    """Model based on SASI-dominated simulations from `Walk et al.,
    PRD 101:123013, 2019 <https://arxiv.org/abs/1910.12971>`_. Data files are
    from the `Garching Supernova Archive`_.
    """
    def __init__(self, * ,progenitor_mass:u.Quantity, direction:int):
        filename = f's{progenitor_mass.value:3.1f}c_3DBH_dir{direction}'
        return super().__init__(filename=filename, metadata=self.metadata)


@RegistryModel(
    progenitor_mass = Parameter(values=(list(range(12, 34)) +
                                 list(range(35, 61, 5)) +
                                 [70, 80, 100, 120]) * u.Msun,
                                desc_values='[12..33, 35..5..60, 70, 80, 100, 120] solMass'
                               ),
    eos = ['HShen', 'LS220']
)
class _OConnor_2013_new(loaders.OConnor_2013):
    """Model based on the black hole formation simulation in `O'Connor & Ott (2013) <https://arxiv.org/abs/1207.1100>`_.
    """
    def __init__(self, eos:str, progenitor_mass:u.Quantity):
        # Load from Parameters
        filename = f'{eos}_timeseries.tar.gz'
        return super().__init__(filename=filename, metadata=self.metadata)

class OConnor_2013(legacy_filename_initialization(_OConnor_2013_new)):
    _config_path='ccsn.OConnor_2013'

    @deprecated('base', 'mass')
    def __init__(self, base=None, mass=None, eos='LS220', *, progenitor_mass=None):
        """
        Parameters
        ----------
        base : str
            Absolute or relative path folder with model data. This argument is deprecated.
        mass: int
            Mass of model progenitor in units Msun. This argument is deprecated.
        """
        if mass:
            progenitor_mass = mass*u.Msun
        if base:
            self.metadata = {'Progenitor mass': progenitor_mass,
                             'EOS': eos}
            filename = f'{base}/{eos}_timeseries.tar.gz'
            return super().__init__(filename=filename, eos=eos, progenitor_mass=progenitor_mass)
        else:
            return super().__init__(eos=eos, progenitor_mass=progenitor_mass)

OConnor_2013.__init__.__doc__=dedent(OConnor_2013.__init__.__doc__)+_OConnor_2013_new.__init__.__doc__
#make sure that only the latest class is present in the models table
all_models.remove(OConnor_2013.__mro__[1])
all_models.add(OConnor_2013)

@deprecated('eos')
@legacy_filename_initialization
@RegistryModel(
    progenitor_mass = [40 * u.Msun],
    eos=['LS220']
)
class OConnor_2015(loaders.OConnor_2015):
    """Model based on the black hole formation simulation in `O'Connor (2015) <https://arxiv.org/abs/1411.7058>`_.
    """
    def __init__(self, progenitor_mass:u.Quantity):
        # Filename is currently the same regardless of parameters
        filename = 'M1_neutrinos.dat'
        return super().__init__(filename, self.metadata)

@deprecated('eos')
@legacy_filename_initialization
@RegistryModel(
    progenitor_mass = Parameter(values=(list(range(16, 27)) + [19.89, 22.39, 30, 33]) * u.Msun,
                                desc_values = '[16..26, 19.89, 22.39, 30, 33] solMass'
                               ),
    eos = ['STOS_B145']
)
class Zha_2021(loaders.Zha_2021):
    """Model based on the hadron-quark phse transition models from `Zha et al. 2021 <https://arxiv.org/abs/2103.02268>`_.
    """
    def _metadata_from_filename(self, filename:str)->dict:
        metadata = {'Progenitor mass': float(os.path.splitext(os.path.basename(filename))[0][1:]) * u.Msun}
        return metadata

    def __init__(self, *, progenitor_mass:u.Quantity):
        filename = f's{progenitor_mass.value:g}.dat'
        return super().__init__(filename, self.metadata)

@deprecated('eos')
@legacy_filename_initialization
@RegistryModel(
    progenitor_mass=Parameter(np.concatenate((np.linspace(9.0, 12.75, 16),
                                              np.linspace(13, 30., 171),
                                              np.linspace(31., 33., 3),
                                              np.linspace(35, 55, 5),
                                              np.linspace(60, 80, 3),
                                              np.linspace(100, 120, 2))) << u.Msun,
                              desc_values='[9..0.25..13, 13..0.1..30, 31..33, 35..5..60, 70, 80, 100, 120] solMass'),
    turbmixing_param= Parameter([1.23, 1.25, 1.27],
                              name='turbmixing_param',
                              label='Turb. mixing param.',
                              description='Turbulent mixing parameter alpha_lambda',
                              ),
    eos=['SFHo']
)
class Warren_2020(loaders.Warren_2020):
    """Model based on simulations from Warren et al., ApJ 898:139, 2020.
    Neutrino fluxes available at https://doi.org/10.5281/zenodo.3667908."""
    # TODO: (For v2.0) Resolve Zenodo issues (Missing files)
    # np.arange with decimal increments can produce floating point errors
    # Though it may be more intuitive to use np.arange, these fp-errors quickly become troublesome
    def _metadata_from_filename(self, filename:str)->dict:
        _, _, turbmixing_param, progenitor_mass = os.path.splitext(os.path.basename(filename))[0].split('_')
        metadata = {'Progenitor mass': float(progenitor_mass[1:]) * u.Msun,
                    'Turb. mixing param.': float(turbmixing_param[1:]),
                    'EOS': 'SFHo'}
        return metadata

    def __init__(self, *, progenitor_mass, turbmixing_param):
        if progenitor_mass.value.is_integer() and progenitor_mass.value <= 30.:
            fname = f'stir_a{turbmixing_param:3.2f}/stir_multimessenger_a{turbmixing_param:3.2f}_m{progenitor_mass.value:.1f}.h5'
        else:
            fname = f'stir_a{turbmixing_param:3.2f}/stir_multimessenger_a{turbmixing_param:3.2f}_m{progenitor_mass.value:g}.h5'
        return super().__init__(fname, self.metadata)

@deprecated('eos','mass')
@legacy_filename_initialization
@RegistryModel(
    rotational_velocity= [0, 1] * u.rad / u.s,
    magnetic_field_exponent= Parameter([0, 12, 13],
                                      label='B_0 Exponent',
                                      description='Exponent of magnetic field (See Eq. 46)'),
    eos=['LS220'],
    progenitor_mass=[20*u.Msun],
    _param_validator = lambda p: (p['rotational_velocity'].value == 1 and p['magnetic_field_exponent'] in (12, 13)) or \
                               (p['rotational_velocity'].value == 0 and p['magnetic_field_exponent'] == 0)
)
class Kuroda_2020(loaders.Kuroda_2020):
    """Model based on simulations from `Kuroda et al. (2020) <https://arxiv.org/abs/2009.07733>`_."""

    def _metadata_from_filename(self, filename:str)->dict:
        _, rotational_velocity, magnetic_field_exponent = re.split('R|B', os.path.splitext(os.path.basename(filename))[0])
        metadata = {
            'Progenitor mass': 20 * u.Msun,
            'Rotational Velocity': int(rotational_velocity[0]),
            'B_0 Exponent': int(magnetic_field_exponent),
            'EOS': 'LS220'
        }
        return metadata

    def __init__(self, mass=None, *, rotational_velocity, magnetic_field_exponent):
        filename = f'LnuR{int(rotational_velocity.value):1d}0B{int(magnetic_field_exponent):02d}.dat'
        return super().__init__(filename, self.metadata)

@legacy_filename_initialization
@RegistryModel(
    progenitor_mass=[9, 10, 12, 13, 14, 15, 16, 19, 25, 60] * u.Msun,
)
class Fornax_2019(loaders.Fornax_2019):
    """Model based on 3D simulations from D. Vartanyan, A. Burrows, D. Radice, M.  A. Skinner and J. Dolence, MNRAS 482(1):351, 2019.
       Data available at https://www.astro.princeton.edu/~burrows/nu-emissions.3d/
    """
    def _metadata_from_filename(self, filename:str)->dict:
        progenitor_mass = os.path.splitext(os.path.basename(filename))[0].split('_')[2]
        metadata = {'Progenitor mass': int(progenitor_mass[:-1]) * u.Msun}
        return metadata

    def __init__(self, cache_flux=False, *, progenitor_mass):
        """
        Parameters
        ----------
        cache_flux : bool
            If true, pre-compute the flux on a fixed angular grid and store the values in a FITS file.
        """
        if progenitor_mass.value == 16:
            filename = f'lum_spec_{int(progenitor_mass.value):d}M_r250.h5'
        else:
            filename = f'lum_spec_{int(progenitor_mass.value):d}M.h5'
        return super().__init__(filename, self.metadata, cache_flux=cache_flux)

@legacy_filename_initialization
@RegistryModel(
    progenitor_mass = Parameter((list(range(12, 24)) + [25, 26, 26.99]) * u.Msun,
                                desc_values='[12..23, 25, 26, 26.99] solMass')
)
class Fornax_2021(loaders.Fornax_2021):
    """Model based on 3D simulations from D. Vartanyan, A. Burrows, D. Radice, M.  A. Skinner and J. Dolence, MNRAS 482(1):351, 2019.
       Data available at https://www.astro.princeton.edu/~burrows/nu-emissions.3d/
        """
    def _metadata_from_filename(self, filename:str)->dict:
        progenitor_mass = os.path.splitext(os.path.basename(filename))[0].split('_')[2]
        metadata = {'Progenitor mass': float(progenitor_mass[:-1]) * u.Msun}
        return metadata

    def __init__(self, progenitor_mass:u.Quantity):
        # Load from Parameters
        if progenitor_mass.value.is_integer():
            filename = f'lum_spec_{int(progenitor_mass.value):2d}M_r10000_dat.h5'
        else:
            filename = f'lum_spec_{progenitor_mass.value:.2f}M_r10000_dat.h5'
        return super().__init__(filename, self.metadata)


_fornax_2022_progenitors = [  '9.0',     '9.25',     '9.5',      '9.75',     '10.0',
                  '10.25',    '10.5',     '10.75',    '11.0',     '11.25',
                  '11.5',     '11.75',    '12.00.bh', '12.03.bh', '12.07.bh',
                  '12.1.bh',  '12.13',    '12.15',    '12.18.bh', '12.20.bh',
                  '12.25',    '12.33.bh', '12.40.bh', '12.45.bh', '12.50.bh',
                  '12.54.bh', '12.60.bh', '12.63',    '12.70',    '12.72.bh',
                  '12.75',    '12.80.bh', '12.85.bh', '12.90.bh', '12.93',
                  '12.97.bh', '13.00.bh', '13.05.bh', '13.11',    '13.25.bh',
                  '13.27.bh', '13.32.bh', '13.40.bh', '13.45',    '13.50.bh',
                  '13.60.bh', '13.75',    '13.82.bh', '13.90.bh', '13.96',
                  '14.01',    '14.13.bh', '14.25.bh', '14.40.bh', '14.41.bh',
                  '14.43',    '14.44.bh', '14.70.bh', '14.87.bh', '15.00.bh',
                  '15.01',    '15.04.bh', '15.05',    '15.38.bh', '16.43',
                  '16.65',    '16.99',    '17.00',    '17.07',    '17.10',
                  '17.40',    '17.48',    '17.50',    '17.51',    '17.83',
                  '18.04',    '18.05',    '18.09',    '18.10',    '18.50',
                  '19.02',    '19.56',    '19.83',    '19.99',    '20.08',
                  '20.09',    '20.18',    '20.37',    '21.00',    '21.68',
                  '22.00',    '22.30',    '22.82',    '23.00',    '23.04',
                  '23.43',    '24.00',    '25.00',    '26.00',    '26.99']

_fornax_2022_masses = [float(p.strip('.bh')) for p in _fornax_2022_progenitors] << u.Msun

@RegistryModel(progenitor_mass = _fornax_2022_masses )
class Fornax_2022(loaders.Fornax_2022):
    """Model based on 2D simulations of 100 progenitors from Tianshu Wang, David Vartanyan, Adam Burrows, and Matthew S.B. Coleman, MNRAS 517:543, 2022.
       Data available at https://www.astro.princeton.edu/~burrows/nu-emissions.2d.large/
        """
    #a mapping of mass to the progenitor
    _mass_to_progenitor = dict(zip(_fornax_2022_masses,_fornax_2022_progenitors))

    def __init__(self, progenitor_mass:u.Quantity):
        progenitor = self._mass_to_progenitor[progenitor_mass]
        self.metadata['Black hole'] = progenitor.endswith('.bh')
        filename = f'lum_spec_{progenitor}_dat.h5'
        return super().__init__(filename, self.metadata)


@RegistryModel(
               _param_validator = lambda p: \
               (p['axion_mass'] == 0 and p['axion_coupling'] == 0) or \
               (p['axion_mass'].to_value('MeV') == 100 and p['axion_coupling'].to_value('1e-10/GeV') in (2,4,10,12,14,16,20)) or \
               (p['axion_mass'].to_value('MeV') == 200 and p['axion_coupling'].to_value('1e-10/GeV') in (2,4,6,8,10,20)),

               axion_mass = Parameter(values=[0, 100, 200]<<u.MeV,
                                      description='Axion mass in units of MeV'),
               axion_coupling = Parameter(values=[0, 2, 4, 6, 8, 10, 12, 14, 16, 20]<<(1e-10/u.GeV),
                                          description='Axion-photon coupling, in units of 1e-10/GeV',
                                          precision=2 #round to 1e-12/u.GeV
                                         ),
               progenitor_mass=[20]*u.Msun
              )
class Mori_2023(loaders.Mori_2023):
    """Model based on 2D simulations with axionlike particles, K. Mori, T.  Takiwaki, K. Kotake and S. Horiuchi, Phys. Rev. D 108:063027, 2023. All models are based on the non-rotating 20 M_sun solar metallicity progenitor model from S.E. Woolsey and A. Heger, Phys. Rep. 442:269, 2007. Data from private communication.
        """
    def __init__(self, axion_mass:u.Quantity, axion_coupling:u.Quantity):
        # Make sure axion coupling is converted to units 1e-10/GeV:
        #axion_coupling = np.round(axion_coupling.to('1e-10/GeV'))

        if axion_mass == 0:
            filename = 't-prof_std.dat'
        else:
            filename = f't-prof_{axion_mass.to_value("MeV"):g}_{axion_coupling.to_value("1e-10/GeV"):g}.dat'

        # PNS mass table, from Mori+ 2023.
        mpns = { (0, 0):    1.78,
                 (100, 2):  1.77,
                 (100, 4):  1.76,
                 (100, 10): 1.77,
                 (100, 12): 1.77,
                 (100, 14): 1.77,
                 (100, 16): 1.77,
                 (100, 20): 1.74,
                 (200, 2):  1.77,
                 (200, 4):  1.76,
                 (200, 6):  1.75,
                 (200, 8):  1.74,
                 (200, 10): 1.73,
                 (200, 20): 1.62 }

        am = int(axion_mass.to_value('MeV')) if isinstance(axion_mass, u.quantity.Quantity) else int(axion_mass)
        ac = int(axion_coupling.to_value('1e-10/GeV')) if isinstance(axion_coupling, u.quantity.Quantity) else int(axion_coupling)

        pns_mass = mpns[(am,ac)]

        # Set the metadata.
        self.metadata['PNS mass'] = pns_mass*u.Msun
        return super().__init__(filename, self.metadata)

@RegistryModel(
    Bfield = ['hydro','L1','L2'],
    direction = ['average','equator','north','south'],
    grav=['A','B',None],
    rotation=[0,90,None],
    _param_validator = lambda p: (p['Bfield'] == 'hydro' and p['grav'] == None and p['rotation'] == None ) or
        (p['Bfield'] == 'L1' and p['grav'] == None and p['rotation'] in [0,90]) or
        (p['Bfield'] == 'L2' and p['rotation'] == None and p['grav'] in ['A','B'])
)
class Bugli_2021(loaders.Bugli_2021):
    """Model based on `Buggli (2021) <https://arxiv.org/abs/2105.00665>`_.
    """
    def __init__(self, *, Bfield:str, direction:str, rotation:int=None, grav:str=None):

        filename = f'{Bfield}_3d_snewpy_{direction}.dat'
        self.metadata['Progenitor mass'] = 35*u.Msun
        self.metadata['EOS'] = 'LS220'
        if Bfield=='L2':
            filename = f'{Bfield}_b12_dipdecay_3d_grav{grav}_snewpy_{direction}.dat'
        if Bfield=='L1':
            filename = f'{Bfield}_b12_3d_{rotation}deg_snewpy_{direction}.dat'
        return super().__init__(filename=filename, metadata=self.metadata)

class SNOwGLoBES:
    """A model that does not inherit from SupernovaModel (yet) and imports a group of SNOwGLoBES files."""

    def __init__(self, tarfilename):
        """
        Parameters
        ----------
        tarfilename: str
            Absolute or relative path to tar archive with SNOwGLoBES files.
        """
        self.tfname = tarfilename
        tf = tarfile.open(self.tfname)

        # For now just pull out the "NoOsc" files.
        datafiles = sorted([f.name for f in tf if '.dat' in f.name])
        noosc = [df for df in datafiles if 'NoOsc' in df]
        noosc.sort(key=len)

        # Loop through the noosc files and pull out the number fluxes.
        self.time = []
        self.energy = None
        self.flux = {}
        self.fmin = 1e99
        self.fmax = -1e99

        for nooscfile in noosc:
            with tf.extractfile(nooscfile) as f:
                logging.debug('Reading {}'.format(nooscfile))
                meta = f.readline()
                metatext = meta.decode('utf-8')
                t = float(metatext.split('TBinMid=')[-1].split('sec')[0])
                dt = float(metatext.split('tBinWidth=')[-1].split('s')[0])
                dE = float(metatext.split('eBinWidth=')[-1].split('MeV')[0])

                data = Table.read(f, format='ascii.commented_header', header_start=-1)
                data.meta['t'] = t
                data.meta['dt'] = dt
                data.meta['dE'] = dE

                self.time.append(t)
                if self.energy is None:
                    self.energy = (data['E(GeV)'].data*1000).tolist()

            for flavor in ['NuE', 'NuMu', 'NuTau', 'aNuE', 'aNuMu', 'aNuTau']:
                if flavor in self.flux:
                    self.flux[flavor].append(data[flavor].data.tolist())
                else:
                    self.flux[flavor] = [data[flavor].data.tolist()]

        # We now have a table with rows=times and columns=energies. Transpose
        # so that rows=energy and cols=time.
        for k, v in self.flux.items():
            self.flux[k] = np.transpose(self.flux[k])
            self.fmin = np.minimum(self.fmin, np.min(self.flux[k]))
            self.fmax = np.maximum(self.fmax, np.max(self.flux[k]))

    def get_fluence(self, t):
        """Return the fluence at a given time t.

        Parameters
        ----------
        t : float
            Time in seconds.

        Returns
        -------
        fluence : dict
            A dictionary giving fluence at time t, keyed by flavor.
        """
        # Get index of closest element in the array
        idx = np.abs(np.asarray(self.time) - t).argmin()

        fluence = {}
        for k, fl in self.flux.items():
            fluence[k] = fl[:,idx]

        return fluence

class Analytic3Species(PinchedModel):
    """An analytical model calculating spectra given total luminosity,
    average energy, and rms or pinch, for each species.
    """

    param = "There are no input files available for this class. Use `doc/scripts/Analytic.py` in the SNEWPY GitHub repo to create a custom input file."

    def get_param_combinations(cls):
        print(cls.param)
        return []

    def __init__(self, filename):
        """
        Parameters
        ----------
        filename : str
            Absolute or relative path to file with model data.
        """

        simtab = Table.read(filename,format='ascii')
        self.filename = filename
        super().__init__(simtab, metadata={})



import os

from snewpy.snowglobes import generate_time_series as NEW_generate_time_series
from snewpy.snowglobes import generate_fluence as NEW_generate_fluence


def generate_time_series(model_path, model_file, model_type, transformation_type, transformation_parameters, d, output_filename, ntbins, deltat):
    print("[INFO] The `generate_time_series` function has been moved to the `snewpy.snowglobes` module.")
    model_path = os.path.join(model_path, model_file)
    if ntbins is None:
        ntbins = 30
    tarfile_path = NEW_generate_time_series(model_path, model_type, transformation_type, d, output_filename, ntbins, deltat)
    return os.path.split(tarfile_path)[1]


def generate_fluence(model_path, model_file, model_type, transformation_type, d, output_filename, tstart=None, tend=None):
    print("[INFO] The `generate_fluence` function has been moved to the `snewpy.snowglobes` module.")
    model_path = os.path.join(model_path, model_file)
    tarfile_path = NEW_generate_fluence(model_path, model_type, transformation_type, d, output_filename, tstart, tend)
    return os.path.split(tarfile_path)[1]



import numpy as np
import astropy.units as u
import pytest
from snewpy.models import ccsn
from snewpy.rate_calculator import RateCalculator

from snewpy import flavor_transformation as ft
pytestmark=pytest.mark.snowglobes

#numbers from SNEWSv2 paper - for approximate checks 
crosscheck_table=[  ('icecube',          320e3),
                    ('wc100kt30prct',4044.8/0.32),
                    ('ar40kt',              2700),
                    ('novaND',                40),
                    ('novaFD',              1900),
                    ('halo1',                  4),
                    ('halo2',                 53)]

#numbers from the SNEWPY SNOwGLoBES interface - for approximate checks 
#of simple rate computation (no smearing)
crosscheck_table_unsmeared=[  ('icecube',          7.2352e6),
                    ('wc100kt30prct',4486.9/0.32),
                    ('ar40kt',              2858.7),
                    ('novaND',                41.7231),
                    ('novaFD',              1947.0768),
                    ('halo1',                  11.5709),
                    ('halo2',                 146.4671)]

@pytest.fixture(autouse=True)
def snmodel():
    return ccsn.Bollig_2016(progenitor_mass=11.2<<u.Msun)

@pytest.fixture(autouse=True)
def rc():
    return RateCalculator()
    
@pytest.fixture
def fluence(snmodel):
    times    = snmodel.get_time()
    energies = np.linspace(0.5,100,201)<<u.MeV
    #get the flux from the model
    flux = snmodel.get_flux(t=times, E=energies, distance=10<<u.kpc, flavor_xform=ft.AdiabaticMSW())
    return flux.integrate('time')

@pytest.mark.parametrize('detector, expected_total',crosscheck_table_unsmeared)
def test_rate_unsmeared(rc, fluence, detector, expected_total):    
    #calculate the event rates
    rates = rc.run(fluence, detector, detector_effects=False)
    #calculate the total IBD event rate 
    N_total = sum([r.integrate_or_sum('energy').array.value for r in rates.values()])
    print(N_total)
    #check the final value
    assert N_total == pytest.approx(expected_total, 0.01)
    
@pytest.mark.parametrize('detector, expected_total',crosscheck_table)
def test_rate_smeared(rc, fluence, detector, expected_total):
    #calculate the event rates
    rates = rc.run(fluence, detector, detector_effects=True)
    #calculate the total IBD event rate 
    N_total = sum([r.integrate_or_sum('energy').array.value for r in rates.values()])
    #check the final value
    assert N_total == pytest.approx(expected_total, 0.1)



import pytest
from pathlib import Path
from snewpy.test._rate_crosscheck_table import rate_table
from snewpy import snowglobes, model_path

from snewpy.models import ccsn
import astropy.units as u

pytestmark=pytest.mark.snowglobes

#get available model parameters from table
param_values = list(rate_table.keys())
#get available detectors from table
detectors = list(list(rate_table.values())[0].keys())

#make sure the model files are loaded
model = ccsn.Bollig_2016
for params in model.get_param_combinations():
    model(**params)
    
def fluence_calculation(model_name,model_file,transform):
    #generating fluence file
    model_file_path = f'{model_path}/{model_name}/{model_file}'
    print(model_file_path)
    return snowglobes.generate_fluence(model_file_path, model_name, transform,d=10)

def rates_calculation(fluence_file):
    tables = snowglobes.simulate(None,fluence_file,detector_input=detectors)
    result = {}
    for det,table in tables.items():
        table = list(table.values())[0] #take the first and the only time bin
        result[det] = table['weighted']['smeared'].values.sum()
    return result

@pytest.mark.parametrize('model_parameters',param_values)
def test_total_rate_equals_table_value(model_parameters):
    fluence_file = fluence_calculation(*model_parameters)
    calculated_rates  = rates_calculation(fluence_file)
    for detector in detectors:
        expected = pytest.approx(rate_table[model_parameters][detector], rel=0.01)
        assert calculated_rates[detector] == expected, f"Crosscheck failed for {detector}"
